import os
import sys
import traceback
from argparse import ArgumentParser

from album.api import Album
from album.argument_parsing import ArgumentParser as AlbumAP
from album.ci.commandline import configure_repo, configure_ssh, zenodo_publish, zenodo_upload, update_index, \
    commit_changes, merge
from album.ci.controller.release_manager import ReleaseManager
from album.runner import album_logging
from album.runner.album_logging import get_active_logger, debug_settings, pop_active_logger

module_logger = get_active_logger


def main():
    """Entry points of `album ci`."""
    ci_parser = create_parser()

    module_logger().info("Starting CI release cycle...")
    args = ci_parser.parse_args()

    album_ci_command = ""
    try:
        album_ci_command = sys.argv[1]  # album command always expected at second position
    except IndexError:
        ci_parser.error("Please provide a valid action!")

    pop_active_logger()  # there will be a new one with the album instance

    # Makes sure album is initialized.
    album_instance = create_album_instance()
    album_instance.load_or_create_collection()

    album_logging.set_loglevel(args.log)

    release_manager = ReleaseManager(album_instance, args.name, args.path, args.src, args.force_retrieve)
    module_logger().debug("Running %s command..." % album_ci_command)

    try:
        args.func(release_manager, args)  # execute entry point function
    except Exception as e:
        _handle_exception(e)
    finally:
        release_manager.close()


def _handle_exception(e):
    get_active_logger().error('album-ci command failed: %s' % str(e))
    get_active_logger().debug(traceback.format_exc())
    sys.exit(e)


def create_album_instance() -> Album:
    return Album.Builder().build()


def create_parser():
    parser = AlbumCIParser()

    parser.create_git_command_parser(
        'configure-repo',
        configure_repo,
        'Configures the git configuration for the catalog repository. Sets an name and an email.'
    )

    p = parser.create_catalog_command_parser(
        'configure-ssh',
        configure_ssh,
        'Configures the git push option for the catalog repository to use the ssh protocol instead of https. '
        'Needs the systems git configuration to be configured for ssh usage!'
    )
    p.add_argument(
        '--ci-project-path',
        required=False,
        type=str,
        help='The project path of the git repository.'
             'E.g. for https://gitlab.com/album-app/album.git the project path is \"album-app/album\". '
             'Usually necessary for using album-ci from within a CI-pipeline.'
             'Allows switching from https to ssh source.'
             'Can be set via environment variable \"CI_PROJECT_PATH\".',
        default=os.getenv('CI_PROJECT_PATH', None)
    )

    parser.create_zenodo_command_parser(
        'publish',
        zenodo_publish,
        'Publishes the corresponding zenodo deposit of a catalog repository deployment branch.'
    )
    p = parser.create_zenodo_command_parser(
        'upload',
        zenodo_upload,
        'Uploads solution of a catalog repository deployment branch to zenodo.'
        'Thereby only allowing a single solution per branch.'
    )
    p.add_argument(
        '--report-file',
        required=False,
        help='Path to a report file. If given, a report of the upload will be created.',
        default="",
        type=str,
    )

    p = parser.create_branch_command_parser(
        'update',
        update_index,
        'Updates the index of the catalog repository to include the solution of a catalog repository deployment branch.'
    )
    p.add_argument(
        '--doi',
        required=False,
        help='Sets the DOI of the solution. Overwrites DOI if already specified in the solution.',
        default="",
        type=str,
    )
    p.add_argument(
        '--deposit-id',
        required=False,
        help='Sets the zenodo deposit_id of the solution.',
        default="",
        type=str,
    )

    parser.create_branch_command_parser(
        'commit',
        commit_changes,
        'Pushes all changes to catalog repository deployment branch to the branch origin.'
    )

    parser.create_pipeline_command_parser(
        'merge',
        merge,
        'Merge the updated catalog index from the given branch to main.'
    )

    return parser.parser


class AlbumCIParser(AlbumAP):
    def __init__(self):
        super().__init__()
        self.parent_parser = self.create_parent_parser()
        self.parser = self.create_parser()
        self.subparsers = self.parser.add_subparsers(title='actions', help='sub-command help')

    @staticmethod
    def create_parent_parser():
        """Parent parser for all subparsers to have the same set of arguments."""
        parent_parser = ArgumentParser(add_help=False)
        # parse logging
        parent_parser.add_argument(
            '--log',
            required=False,
            help='Logging level for your album-ci command. Choose between %s' %
                 ", ".join([loglevel.name for loglevel in album_logging.LogLevel]),
            default=album_logging.LogLevel(debug_settings()),
            type=(lambda choice: album_logging.to_loglevel(choice)),
        )
        return parent_parser

    def create_parser(self):
        """Creates the main parser for the album framework."""
        parser = ArgumentParser(
            add_help=True,
            description=
            'album-catalog-admin for managing catalog-deploy requests to a catalog via commandline or CI',
            parents=[self.parent_parser])
        return parser

    def create_catalog_command_parser(self, command_name, command_function, command_help):
        """Creates a subparser with all necessary arguments for the catalog deployment management"""
        parser = self.subparsers.add_parser(command_name, help=command_help, parents=[self.parent_parser])
        parser.set_defaults(func=command_function)
        parser.add_argument('name', type=str, help='Name of the catalog.')
        parser.add_argument('path', type=str, help='Path of the catalog repository. Can be an empty folder.')
        parser.add_argument('src', type=str, help='Source of the catalog. Usually a git repository link.')

        parser.add_argument(
            '--force-retrieve',
            required=False,
            help='If True, download path for the catalog will be force emptied before retrieving the catalog.',
            default=False,
            type=bool,
        )

        return parser

    def create_git_command_parser(self, command_name, command_function, command_help):
        parser = self.create_catalog_command_parser(command_name, command_function, command_help)
        parser.add_argument(
            '--ci-user-name',
            required=False,
            type=str,
            help='Name to use for all ci operations. '
                 'Can be set via environment variable \"CI_USER_NAME\".'
                 'If neither is given, system is required to be proper configured!',
            default=os.getenv('CI_USER_NAME', None)
        )
        parser.add_argument(
            '--ci-user-email',
            required=False,
            type=str,
            help='Email to use for all ci operations. '
                 'Can be set via environment variable \"CI_USER_EMAIL\".'
                 'If neither is given, system is required to be proper configured!',
            default=os.getenv('CI_USER_EMAIL', None)
        )

        return parser

    def create_branch_command_parser(self, command_name, command_function, command_help):
        parser = self.create_git_command_parser(command_name, command_function, command_help)
        parser.add_argument(
            '--branch-name',
            required=True,
            type=str,
            help='The branch name of the solution to publish.'
        )
        return parser

    def create_zenodo_command_parser(self, command_name, command_function, command_help):
        parser = self.create_branch_command_parser(command_name, command_function, command_help)
        parser.add_argument(
            '--zenodo-base-url',
            required=False,
            type=str,
            help='The base URL where to upload to. Either \"https://sandbox.zenodo.org\" or \"https://zenodo.org\"'
                 'Can be set via environment variable \"ZENODO_BASE_URL\".',
            default=os.getenv('ZENODO_BASE_URL', None)
        )
        parser.add_argument(
            '--zenodo-access-token',
            required=False,
            type=str,
            help='The access token representing the zenodo user account to use. For help see zenodo.org!'
                 'Can be set via environment variable \"ZENODO_ACCESS_TOKEN\".',
            default=os.getenv('ZENODO_ACCESS_TOKEN', None)
        )
        return parser

    def create_pipeline_command_parser(self, command_name, command_function, command_help):
        parser = self.create_branch_command_parser(command_name, command_function, command_help)
        parser.add_argument(
            '--dry-run',
            required=False,
            help='Dry-run option.'
                 ' If this argument is added, commits will not be pushed, only information is shown.',
            action='store_true'
        )
        parser.add_argument(
            '--push-option',
            required=False,
            help='Push options for the catalog repository.',
            default=None,
            nargs="+"
        )
        return parser



import json
import pkgutil

from jsonschema import validate

from album.core.api.controller.controller import IAlbumController
from album.core.api.controller.migration_manager import IMigrationManager
from album.core.api.model.catalog import ICatalog
from album.core.api.model.collection_index import ICollectionIndex
from album.core.model.catalog_index import CatalogIndex
from album.core.model.collection_index import CollectionIndex
from album.runner import album_logging

module_logger = album_logging.get_active_logger


class MigrationManager(IMigrationManager):

    def __init__(self, album: IAlbumController):
        self.schema_solution = None
        self.album = album

    def migrate_collection_index(self, collection_index: ICollectionIndex, initial_version):
        self.migrate_catalog_collection_db(
            collection_index.get_path(),
            initial_version,  # current version
            CollectionIndex.version  # current framework target version
        )

    def _create_catalog_index(self, catalog: ICatalog, initial_version) -> None:
        catalog.load_index()
        self.migrate_catalog_index_db(
            catalog.index().get_path(),
            initial_version,  # current version
            CatalogIndex.version  # current framework target version
        )

    def migrate_catalog_collection_db(self, collection_index_path, curr_version, target_version):
        if curr_version != target_version:
            # todo: execute catalog_collection SQL migration scripts if necessary!
            raise NotImplementedError(
                "Cannot migrate collection from version \"%s\" to version \"%s\"!" % (curr_version, target_version)
            )
        return collection_index_path

    def migrate_catalog_index_db(self, catalog_index_path, curr_version, target_version):
        if curr_version != target_version:
            # todo: execute catalog index SQL migration scripts if necessary!
            raise NotImplementedError(
                "Cannot migrate collection from version %s to version %s." % (curr_version, target_version)
            )
        return catalog_index_path

    def load_index(self, catalog: ICatalog):
        catalog.update_index_cache()

        self._create_catalog_index(catalog, CatalogIndex.version)
        self.album.catalogs().set_version(catalog)

    def refresh_index(self, catalog: ICatalog) -> bool:
        """Routine to refresh the catalog index. Downloads or copies the index_file."""
        if catalog.update_index_cache_if_possible():
            self._create_catalog_index(catalog, CatalogIndex.version)
            return True
        return False

    def validate_solution_attrs(self, attrs):
        self._load_solution_schema()
        validate(attrs, self.schema_solution)

    def _load_solution_schema(self):
        if not self.schema_solution:
            data = pkgutil.get_data('album.core.schema', 'solution_schema_0.json')
            self.schema_solution = json.loads(data)



from pathlib import Path
from typing import Generator

from git import Repo

from album.api import Album
from album.ci.controller.zenodo_manager import ZenodoManager
from album.ci.utils.continuous_integration import get_ssh_url, create_report
from album.core.model.catalog import Catalog, get_index_url
from album.core.utils.export.changelog import get_changelog_file_name
from album.core.utils.operations.file_operations import get_dict_from_yml, write_dict_to_yml, get_dict_entry
from album.core.utils.operations.git_operations import checkout_branch, add_files_commit_and_push, \
    retrieve_files_from_head, configure_git
from album.core.utils.operations.resolve_operations import get_zip_name, get_zip_name_prefix, dict_to_coordinates
from album.core.utils.operations.url_operations import download_resource, is_downloadable
from album.runner import album_logging
from album.runner.core.api.model.coordinates import ICoordinates

module_logger = album_logging.get_active_logger


class ReleaseManager:
    """Class for handling a catalog as administrator."""

    configuration = None

    def __init__(self, album_instance: Album, catalog_name, catalog_path, catalog_src, force_retrieve):
        self.catalog_name = catalog_name
        self.catalog_path = catalog_path
        self.catalog_src = catalog_src

        self.configuration = album_instance.configuration()
        if not self.configuration.is_setup():
            self.configuration.setup()

        self.catalog = Catalog(None, name=self.catalog_name, path=catalog_path, src=self.catalog_src)
        self.force_retrieve = force_retrieve
        self.album_instance = album_instance

    def _open_repo(self) -> Generator[Repo, None, None]:
        repo = self.catalog.retrieve_catalog(force_retrieve=self.force_retrieve, update=False)

        return repo

    def close(self):
        if self.catalog:
            self.catalog.dispose()

    def configure_repo(self, user_name, user_email):
        module_logger().info("Configuring repository using:\n\tusername:\t%s\n\temail:\t%s" % (user_name, user_email))
        with self._open_repo() as repo:
            configure_git(repo, user_email, user_name)

    def configure_ssh(self, project_path):
        module_logger().info("Configure ssh protocol for repository %s" % project_path)
        with self._open_repo() as repo:
            if not repo.remote().url.startswith("git"):
                repo.remote().set_url(get_ssh_url(project_path, self.catalog_src))

    def _get_zip_path(self, coordinates: ICoordinates):
        zip_name = get_zip_name(coordinates)
        return self.configuration.get_solution_path_suffix(coordinates).joinpath(zip_name)

    def _get_docker_path(self, coordinates: ICoordinates):
        docker_name = "Dockerfile"
        return self.configuration.get_solution_path_suffix(coordinates).joinpath(docker_name)

    def _get_changelog_path(self, coordinates: ICoordinates):
        changelog_name = get_changelog_file_name()
        return self.configuration.get_solution_path_suffix(coordinates).joinpath(changelog_name)

    def _get_documentation_paths(self, coordinates: ICoordinates, yml_dict: dict):
        solution_path_suffix = self.configuration.get_solution_path_suffix(coordinates)
        documentation_paths = []
        if "documentation" in yml_dict.keys():
            documentation_list = yml_dict["documentation"]
            if not isinstance(documentation_list, list):
                documentation_list = [documentation_list]

            for documentation_name in documentation_list:
                documentation_paths.append(solution_path_suffix.joinpath(documentation_name))
        return documentation_paths

    @staticmethod
    def _get_yml_dict(head):
        yml_file_path = retrieve_files_from_head(head, '[a-zA-Z0-9]*.yml')[0]
        yml_dict = get_dict_from_yml(yml_file_path)

        return [yml_dict, yml_file_path]

    def zenodo_publish(self, branch_name, zenodo_base_url, zenodo_access_token):
        zenodo_base_url, zenodo_access_token = self.__prepare_zenodo_arguments(zenodo_base_url, zenodo_access_token)

        zenodo_manager = self._get_zenodo_manager(zenodo_access_token, zenodo_base_url)

        with self._open_repo() as repo:
            head = checkout_branch(repo, branch_name)

            # get the yml file to release from current branch
            yml_dict, yml_file_path = self._get_yml_dict(head)

        # checkout files
        zip_name = self._get_zip_path(dict_to_coordinates(yml_dict)).name
        docker_name = self._get_docker_path(dict_to_coordinates(yml_dict)).name

        # query deposit
        deposit_name = get_zip_name_prefix(dict_to_coordinates(yml_dict))
        deposit_id = get_dict_entry(yml_dict, "deposit_id", allow_none=False)

        module_logger().info("Get unpublished deposit with deposit id %s..." % deposit_id)
        deposit = zenodo_manager.zenodo_get_unpublished_deposit_by_id(
            deposit_id, deposit_name, expected_files=[zip_name, docker_name]
        )

        # publish to zenodo
        deposit.publish()
        module_logger().info("Published unpublished deposit with deposit id %s..." % deposit_id)

    def zenodo_upload(self, branch_name, zenodo_base_url, zenodo_access_token, report_file):
        zenodo_base_url, zenodo_access_token = self.__prepare_zenodo_arguments(zenodo_base_url, zenodo_access_token)

        zenodo_manager = self._get_zenodo_manager(zenodo_access_token, zenodo_base_url)
        with self._open_repo() as repo:
            head = checkout_branch(repo, branch_name)

            # get the yml file to release
            yml_dict, yml_file_path = self._get_yml_dict(head)

        # get metadata from yml_file
        try:
            deposit_id = yml_dict["deposit_id"]
        except KeyError:
            deposit_id = None
        coordinates = dict_to_coordinates(yml_dict)

        # do not upload SNAPSHOT versions
        if coordinates.version().endswith("SNAPSHOT"):
            module_logger().info("Will not upload SNAPSHOT version to zenodo! Skipping...")
            return

        # extract deposit name
        deposit_name = get_zip_name_prefix(coordinates)

        # get the solution zip to release
        zip_path = self._get_zip_path(dict_to_coordinates(yml_dict))
        solution_zip = retrieve_files_from_head(head, str(zip_path), option="startswith")[0]

        # get the docker file
        docker_path = self._get_docker_path(dict_to_coordinates(yml_dict))
        docker_file = retrieve_files_from_head(head, str(docker_path), option="startswith")[0]

        # get the changelog file
        changelog_path = self._get_changelog_path(dict_to_coordinates(yml_dict))
        changelog_file = retrieve_files_from_head(head, str(changelog_path), option="startswith")[0]

        # get the documentation files (if any)
        documentation_paths = self._get_documentation_paths(dict_to_coordinates(yml_dict), yml_dict)
        documentation_files = [
            retrieve_files_from_head(head, str(documentation_path), option="startswith")[0]
            for documentation_path in documentation_paths
        ]

        # get the release deposit. Either a new one or an existing one to perform an update on
        deposit = zenodo_manager.zenodo_get_deposit(
            deposit_name, deposit_id, expected_files=[solution_zip, docker_file, changelog_file]
        )
        module_logger().info("Deposit %s successfully retrieved..." % deposit.id)

        # include doi and ID in yml
        yml_dict["doi"] = deposit.metadata.prereserve_doi["doi"]
        yml_dict["deposit_id"] = deposit.id
        write_dict_to_yml(yml_file_path, yml_dict)

        # zenodo upload files but not publish
        deposit = zenodo_manager.zenodo_upload(deposit, solution_zip)
        zenodo_manager.zenodo_upload(deposit, docker_file)
        zenodo_manager.zenodo_upload(deposit, changelog_file)
        for documentation_file in documentation_files:
            zenodo_manager.zenodo_upload(deposit, documentation_file)
        module_logger().info("Deposit %s successfully retrieved..." % deposit_id)

        if report_file:
            report_file = create_report(report_file, {"DOI": yml_dict["doi"], "DEPOSIT_ID": yml_dict["deposit_id"]})
            module_logger().info("Created report file under %s" % str(report_file))

    @staticmethod
    def __prepare_zenodo_arguments(zenodo_base_url: str, zenodo_access_token: str):
        if zenodo_base_url is None or zenodo_access_token is None:
            raise RuntimeError('Zenodo base URL or Zenodo access token invalid! '
                               'See https://zenodo.org/ documentation for information!')

        zenodo_base_url = zenodo_base_url.strip()
        zenodo_access_token = zenodo_access_token.strip()

        if zenodo_base_url == "" or zenodo_access_token == "":
            raise RuntimeError("Empty zenodo base_url or zenodo_access_token!")

        module_logger().info("Using zenodo URL: %s" % zenodo_base_url)

        return zenodo_base_url, zenodo_access_token

    def update_index(self, branch_name, doi, deposit_id):
        with self._open_repo() as repo:
            head = checkout_branch(repo, branch_name)

            # always use remote index
            index_url, index_meta = get_index_url(self.catalog_src)

            if is_downloadable(index_url):
                download_resource(index_url, self.catalog.index_path())
            else:
                module_logger().warning("Index not downloadable! Using Index in merge request branch!")

            self.album_instance.load_catalog_index(self.catalog)

            yml_dict, yml_file_path = self._get_yml_dict(head)

            # update doi and deposit_id if given
            log_message_doi = "without a doi"
            if doi:
                yml_dict["doi"] = doi
                log_message_doi = "with doi %s" % doi

            log_message_deposit = "without deposit"
            if deposit_id:
                yml_dict["deposit_id"] = deposit_id
                log_message_deposit = "in zenodo deposit %s" % deposit_id

            if doi or deposit_id:
                write_dict_to_yml(yml_file_path, yml_dict)

        module_logger().info(
            "Update index to include solution from branch %s %s %s!" %
            (branch_name, log_message_doi, log_message_deposit)
        )

        self.catalog.index().update(dict_to_coordinates(yml_dict), yml_dict)
        self.catalog.index().save()
        self.catalog.index().export(self.catalog.solution_list_path())

    def commit_changes(self, branch_name, ci_user_name, ci_user_email):
        with self._open_repo() as repo:
            head = checkout_branch(repo, branch_name)

            yml_dict, yml_file = self._get_yml_dict(head)

            # get the solution zip to release
            zip_path = self._get_zip_path(dict_to_coordinates(yml_dict))
            solution_zip = retrieve_files_from_head(head, str(zip_path), option="startswith")[0]

            # get the docker file
            docker_path = self._get_docker_path(dict_to_coordinates(yml_dict))
            docker_file = retrieve_files_from_head(head, str(docker_path), option="startswith")[0]

            # get the changelog file
            changelog_path = self._get_changelog_path(dict_to_coordinates(yml_dict))
            changelog_file = retrieve_files_from_head(head, str(changelog_path), option="startswith")[0]

            # get the documentation files (if any)
            documentation_paths = self._get_documentation_paths(dict_to_coordinates(yml_dict), yml_dict)
            documentation_files = [
                retrieve_files_from_head(head, str(documentation_path), option="startswith")[0]
                for documentation_path in documentation_paths
            ]

            commit_files = [yml_file, solution_zip, docker_file, changelog_file] + documentation_files
            if not all([Path(f).is_file() for f in commit_files]):
                raise FileNotFoundError("Invalid deploy request or broken catalog repository!")

            commit_msg = "Prepared branch \"%s\" for merging." % branch_name

            add_files_commit_and_push(
                head,
                commit_files,
                commit_msg,
                push=False,
                username=ci_user_name,
                email=ci_user_email
            )

        return True

    def merge(self, branch_name, dry_run, push_option, ci_user_name, ci_user_email):
        with self._open_repo() as repo:
            head = checkout_branch(repo, branch_name)

            commit_files = [
                self.catalog.solution_list_path(), self.catalog.index_path()
            ]
            if not all([Path(f).is_file() for f in commit_files]):
                raise FileNotFoundError("Invalid deploy request or broken catalog repository!")

            commit_msg = "Updated index."

            module_logger().info("Prepare merging...")

            add_files_commit_and_push(
                head,
                commit_files,
                commit_msg,
                push=not dry_run,
                push_option_list=push_option,
                username=ci_user_name,
                email=ci_user_email
            )

    def _get_zenodo_manager(self, zenodo_access_token, zenodo_base_url):
        # TODO in case one wants to reuse the zenodo manager, this needs to be smarter, but be aware that another call
        #  to this method might have different parameters
        return ZenodoManager(zenodo_base_url, zenodo_access_token)



from distutils import util
from pathlib import Path
from typing import Optional

from flask import Flask, request
from werkzeug.exceptions import abort

import album
from album.api import Album
from album.core.api.controller.task_manager import ITaskManager
from album.core.controller.task_manager import TaskManager
from album.core.model.default_values import DefaultValues
from album.core.model.task import Task
from album.runner import album_logging
from album.runner.core.model.coordinates import Coordinates

module_logger = album_logging.get_active_logger


class AlbumServer:
    port = DefaultValues.server_port.value
    host = DefaultValues.server_port.value

    app = None

    def __init__(self, port: int, host: str = None):
        self.port = port
        self.host = host
        self.album_instance: Optional[Album] = None
        self._task_manager = None

    def setup(self, album_instance: Album):
        self.album_instance = album_instance
        self._task_manager = TaskManager()

    def start(self, test_config=None):
        module_logger().info('Starting server...')
        self.init_server(test_config)
        self.app.run(port=self.port, host=self.host)

    def init_server(self, test_config=None):
        self.app = Flask(__name__, instance_relative_config=True)
        if test_config is not None:
            self.app.config.update(test_config)
        self._set_routes()
        return self.app

    def task_manager(self) -> ITaskManager:
        return self._task_manager

    def _set_routes(self):
        @self.app.route("/")
        def get_version():
            return {
                "version": album.core.__version__,
                "author": album.core.__author__,
                "email": album.core.__email__
            }

        @self.app.route("/config")
        def get_config():
            return {
                "cache_base": str(self.album_instance.configuration().base_cache_path()),
                "cache_tmp_internal": str(self.album_instance.configuration().cache_path_tmp_internal()),
                "cache_tmp_user": str(self.album_instance.configuration().cache_path_tmp_user()),
                "cache_apps": str(self.album_instance.configuration().cache_path_app()),
                "cache_downloads": str(self.album_instance.configuration().cache_path_download())
            }

        @self.app.route("/index")
        def get_index():
            index_dict = self.album_instance.get_index_as_dict()
            return index_dict

        @self.app.route("/catalogs")
        def get_catalogs():
            return self.album_instance.get_catalogs_as_dict()

        @self.app.route("/recently-launched")
        def get_recently_launched_solutions():
            solutions = []
            for solution in self.album_instance.get_collection_index().get_recently_launched_solutions():
                solutions.append({
                    'setup': solution.setup(),
                    'internal': solution.internal()
                })
            return {'solutions': solutions}

        @self.app.route("/recently-installed")
        def get_recently_installed_solutions():
            solutions = []
            for solution in self.album_instance.get_collection_index().get_recently_installed_solutions():
                solutions.append({
                    'setup': solution.setup(),
                    'internal': solution.internal()
                })
            return {'solutions': solutions}

        @self.app.route('/run/<group>/<name>/<version>', defaults={'catalog': None})
        @self.app.route('/run/<catalog>/<group>/<name>/<version>')
        def run(catalog, group, name, version):
            args = self._get_arguments(request.args)
            task = self._run_solution_method_async(catalog, Coordinates(group, name, version),
                                                   self.album_instance.run,
                                                   [args, False])
            return {"id": task.id(), "msg": "process started"}

        @self.app.route('/install/<group>/<name>/<version>', defaults={'catalog': None})
        @self.app.route('/install/<catalog>/<group>/<name>/<version>')
        def install(catalog, group, name, version):
            task = self._run_solution_method_async(
                catalog,
                Coordinates(group, name, version),
                self.album_instance.install
            )
            return {"id": task.id(), "msg": "process started"}

        @self.app.route('/uninstall/<group>/<name>/<version>', defaults={'catalog': None})
        @self.app.route('/uninstall/<catalog>/<group>/<name>/<version>')
        def uninstall(catalog, group, name, version):
            task = self._run_solution_method_async(
                catalog,
                Coordinates(group, name, version),
                self.album_instance.uninstall
            )
            return {"id": task.id(), "msg": "process started"}

        @self.app.route('/test/<group>/<name>/<version>', defaults={'catalog': None})
        @self.app.route('/test/<catalog>/<group>/<name>/<version>')
        def test(catalog, group, name, version):
            task = self._run_solution_method_async(catalog, Coordinates(group, name, version),
                                                   self.album_instance.test)
            return {"id": task.id(), "msg": "process started"}

        @self.app.route('/deploy')
        def deploy():
            solution_path = request.args.get("path")
            catalog_name = request.args.get("catalog_name")
            if solution_path is None:
                abort(404, description=f"`path` argument missing")
            if not Path(solution_path).exists():
                abort(404, description=f"Solution not found: {solution_path}")
            if catalog_name is None:
                abort(404, description=f"`catalog_name` argument missing")
            dryrun = bool(util.strtobool(request.args.get("dryrun", default="false")))
            trigger_pipeline = False
            task = Task()
            task._args = (solution_path, catalog_name, dryrun, trigger_pipeline)
            task._method = self.album_instance.deploy
            self.task_manager().register_task(task)
            return {"id": task.id(), "msg": "process started"}

        @self.app.route('/clone/<group>/<name>/<version>', defaults={'catalog': None})
        @self.app.route('/clone/<catalog>/<group>/<name>/<version>')
        def clone_solution(catalog, group, name, version):
            target_dir = request.args.get("target_dir")
            new_name = request.args.get("name")
            if target_dir is None:
                abort(404, description=f"`target_dir` argument missing")
            if new_name is None:
                abort(404, description=f"`name` argument missing")
            args = [target_dir, new_name]
            task = self._run_solution_method_async(
                catalog,
                Coordinates(group, name, version),
                self.album_instance.clone, args,
                resolve=False
            )
            return {"id": task.id(), "msg": "process started"}

        @self.app.route('/clone')
        def clone_solution_py_path():
            target_dir = request.args.get("target_dir")
            name = request.args.get("name")
            path = request.args.get("path")
            if target_dir is None:
                abort(404, description=f"`target_dir` argument missing")
            if name is None:
                abort(404, description=f"`name` argument missing")
            if path is None:
                abort(404, description=f"`path` argument missing")
            args = [target_dir, name]
            task = Task()
            task_args = [path]
            if args:
                for arg in args:
                    task_args.append(arg)
            task._args = tuple(task_args)
            task._method = self.album_instance.clone
            self.task_manager().register_task(task)
            return {"id": task.id(), "msg": "process started"}

        @self.app.route('/clone/<template_name>')
        def clone_catalog(template_name):
            target_dir = request.args.get("target_dir")
            name = request.args.get("name")
            if target_dir is None:
                abort(404, description=f"`target_dir` argument missing")
            if name is None:
                abort(404, description=f"`name` argument missing")
            task = Task()
            task._args = tuple([template_name, target_dir, name])
            task._method = self.album_instance.clone
            self.task_manager().register_task(task)
            return {"id": task.id(), "msg": "process started"}

        @self.app.route('/status/<catalog>/<group>/<name>/<version>')
        def status_solution(catalog, group, name, version):
            try:
                catalog_id = self.album_instance.get_catalog_by_name(catalog).catalog_id()
                installed = self.album_instance.get_collection_index().is_installed(
                    catalog_id,
                    Coordinates(group, name, version)
                )
                return {
                    "installed": installed
                }
            except LookupError as e:
                abort(404, description=f"Solution not found")

        @self.app.route('/status/<task_id>')
        def status_task(task_id):
            task = self.task_manager().get_task(str(task_id))
            if task is None:
                abort(404, description=f"Task not found with id {task_id}")
            return self.task_manager().get_status(task)

        @self.app.route('/add-catalog')
        def add_catalog():
            url = request.args.get("src")
            catalog = self.album_instance.add_catalog(url)
            catalog_id = catalog.catalog_id()
            return {"catalog_id": catalog_id}

        @self.app.route('/remove-catalog')
        def remove_catalog():
            url = request.args.get("src", default=None)
            name = request.args.get("name", default=None)
            if name is None:
                self.album_instance.remove_catalog_by_src(url)
            else:
                self.album_instance.remove_catalog_by_name(name)
            return {}

        @self.app.route('/upgrade')
        def upgrade():
            src = request.args.get("src", default=None)
            name = request.args.get("name", default=None)
            dry_run = request.args.get("dry_run", default=False)
            if name is None and src is None:
                res = self.album_instance.upgrade(dry_run=dry_run)
            elif name is None:
                catalog = self.album_instance.get_catalog_by_src(src)
                res = self.album_instance.upgrade(catalog_name=catalog.name(), dry_run=dry_run)
            else:
                res = self.album_instance.upgrade(catalog_name=name, dry_run=dry_run)
            r = {}
            for catalog_name in res:
                r[catalog_name] = res[catalog_name].as_dict()
            return r

        @self.app.route('/update')
        def update():
            src = request.args.get("src", default=None)
            name = request.args.get("name", default=None)
            if name is None and src is None:
                self.album_instance.update()
            elif name is None:
                catalog = self.album_instance.get_catalog_by_src(src)
                self.album_instance.update(catalog.name())
            else:
                self.album_instance.update(name)
            return {}

        @self.app.route('/search/<keywords>')
        def search(keywords):
            return self.album_instance.search(keywords)

        @self.app.route('/shutdown', methods=['GET'])
        def shutdown():
            func = request.environ.get('werkzeug.server.shutdown')
            if func is None:
                raise RuntimeError('Not running with the Werkzeug Server')
            func()
            return 'Server shutting down...'

    def _run_solution_method_async(self, catalog, group_name_version: Coordinates, method, args=None, resolve=True):
        task = Task()
        if catalog is None:
            solution = str(group_name_version)
        else:
            solution = ":".join([catalog, str(group_name_version)])
        if resolve:
            solution = self.album_instance.resolve(solution)
        task_args = [solution]
        if args:
            for arg in args:
                task_args.append(arg)
        task._args = tuple(task_args)
        task._method = method
        self.task_manager().register_task(task)
        return task

    @staticmethod
    def _get_arguments(args_json):
        command_args = [""]
        if args_json:
            for key in args_json:
                command_args.append(f"--{key}")
                command_args.append(str(args_json[key]))
        return command_args



import tempfile
from datetime import datetime
from pathlib import Path

from git import Repo

from album.core.api.controller.controller import IAlbumController
from album.core.api.controller.deploy_manager import IDeployManager
from album.core.api.model.catalog import ICatalog
from album.core.model.default_values import DefaultValues
from album.core.utils.export.changelog import create_changelog_file, \
    process_changelog_file
from album.core.utils.export.docker import create_docker_file
from album.core.utils.operations.dict_operations import get_dict_entries_from_attribute_path
from album.core.utils.operations.file_operations import copy, write_dict_to_yml, zip_folder, zip_paths, force_remove, \
    folder_empty
from album.core.utils.operations.git_operations import create_new_head, add_files_commit_and_push, \
    retrieve_default_mr_push_options
from album.core.utils.operations.solution_operations import get_deploy_dict
from album.runner import album_logging
from album.runner.core.api.model.solution import ISolution
from album.runner.core.model.solution import Solution

module_logger = album_logging.get_active_logger


class DeployManager(IDeployManager):

    def __init__(self, album: IAlbumController):
        self.album = album

    def deploy(self, deploy_path, catalog_name: str, dry_run: bool, push_options=None, git_email: str = None,
               git_name: str = None,
               force_deploy: bool = False, changelog: str = None):

        if dry_run:
            module_logger().info('Pretending to deploy %s to %s...' % (deploy_path, catalog_name))
        else:
            module_logger().info('Deploying %s to %s...' % (deploy_path, catalog_name))

        deploy_path = Path(deploy_path)

        if deploy_path.is_dir():
            path_to_solution = deploy_path.joinpath(DefaultValues.solution_default_name.value)
        else:
            path_to_solution = deploy_path

        active_solution = self.album.state_manager().load(path_to_solution)
        active_solution.setup().changelog = changelog

        # case catalog given
        if catalog_name:
            catalog = self.album.catalogs().get_by_name(catalog_name)

        # case catalog in solution file specified
        # TODO: discuss this
        elif active_solution.setup().deploy and active_solution.setup().deploy["catalog"]:
            catalog = self.album.catalogs().get_by_src(
                active_solution.setup().deploy["catalog"]["src"]
            )

        # case no catalog given
        else:
            raise RuntimeError("No catalog specified for deployment!")

        self.album.migration_manager().load_index(catalog)

        process_changelog_file(catalog, active_solution, deploy_path)

        if catalog.is_local():
            self._deploy_to_local_catalog(catalog, active_solution, deploy_path, dry_run, force_deploy)
        else:
            self._deploy_to_remote_catalog(
                catalog, active_solution, deploy_path, dry_run, push_options, git_email, git_name
            )

        if dry_run:
            module_logger().info('Successfully pretended to deploy %s to %s.' % (deploy_path, catalog_name))
        else:
            module_logger().info('Successfully deployed %s to %s.' % (deploy_path, catalog_name))

    def _deploy_to_remote_catalog(self, catalog: ICatalog, active_solution: ISolution, deploy_path, dry_run,
                                  push_options,
                                  git_email=None, git_name=None):
        """Routine to deploy to a remote catalog."""
        dl_path = self.get_download_path(catalog)

        with catalog.retrieve_catalog(dl_path, force_retrieve=True) as repo:
            catalog_local_src = repo.working_tree_dir

            if not repo:
                raise FileNotFoundError("Catalog repository not found. Did the download of the catalog fail?")

            # include files/folders in catalog
            solution_zip, exports = self._deploy_routine_in_local_src(
                catalog, catalog_local_src, active_solution, deploy_path
            )

            # build merge request files
            mr_files = [solution_zip] + exports

            if not push_options:
                push_options = retrieve_default_mr_push_options(catalog.src())

            self._create_merge_request(active_solution, repo, mr_files, dry_run, push_options, git_email, git_name)

    def get_download_path(self, catalog: ICatalog):
        return Path(self.album.configuration().cache_path_download()).joinpath(catalog.name())

    def _deploy_to_local_catalog(self, catalog: ICatalog, active_solution: ISolution, deploy_path, dry_run: bool,
                                 force_deploy: bool):
        """Routine to deploy to a local catalog."""
        # check for cache catalog only
        if catalog.is_cache():
            raise RuntimeError("Cannot deploy to catalog only used for caching! Aborting...")

        # set src to catalog.src, as we know this is a (network) path, not a remote link
        catalog_local_src = catalog.src()
        active_solution.setup()["timestamp"] = datetime.strftime(datetime.now(), '%Y-%m-%dT%H:%M:%S.%f')

        # check if catalog_local_src is empty
        catalog_local_src_solution_path = self._get_absolute_prefix_path(
            catalog, catalog_local_src, active_solution
        )
        if catalog_local_src_solution_path.exists() and not folder_empty(catalog_local_src_solution_path):
            if force_deploy:
                force_remove(catalog_local_src_solution_path)
            else:
                raise RuntimeError("The deploy target folder is not empty! Enable --force-deploy to continue!")

        # update the index
        if not dry_run:
            catalog.add(active_solution, force_overwrite=force_deploy)
        else:
            module_logger().info("Would add the solution %s to index..." % active_solution.coordinates().name())

        # include files/folders in catalog
        zip_path, export = self._deploy_routine_in_local_src(catalog, catalog_local_src, active_solution,
                                                             deploy_path)

        # copy to source
        if not dry_run:
            try:
                catalog.copy_index_from_cache_to_src()
            except Exception:
                module_logger().error("Copying index to src failed! Rolling back deployment...")
                zip_path.unlink()
                for e in export:
                    e.unlink()
                raise OSError("Deploy failed!")

            # refresh the local index of the catalog
            self.album.migration_manager().refresh_index(catalog)
        else:
            module_logger().info(
                "Would copy the index to src %s to %s..." % (catalog.index_path(), catalog.src())
            )
            module_logger().info(
                "Would refresh the index from src"
            )

    def _deploy_routine_in_local_src(self, catalog: ICatalog, catalog_local_src, active_solution, deploy_path):
        """Performs all routines a deploy process needs to do locally.

        Returns:
            solution zip file and additional attachments.

        """
        solution_zip = self._copy_and_zip(catalog_local_src, active_solution, deploy_path)
        exports = self._attach_exports(catalog, catalog_local_src, active_solution, deploy_path)

        return solution_zip, exports

    @staticmethod
    def retrieve_head_name(active_solution: ISolution):
        """Retrieves the branch (head) name for the merge request of the solution file."""
        coordinates = active_solution.coordinates()
        return "_".join([coordinates.group(), coordinates.name(), coordinates.version()])

    @staticmethod
    def _create_merge_request(active_solution: ISolution, repo: Repo, file_paths, dry_run=False, push_option=None,
                              email=None, username=None):
        """Creates a merge request to the catalog repository for the album object.

        Commits first the files given in the call, but will not include anything else than that.

        Args:
            file_paths:
                A list of files to include in the merge request. Can be relative to
                the catalog repository path or absolute.
            dry_run:
                Option to only show what would happen. No Merge request will be created.
            push_option:
                Push options for the repository.
            username:
                The git user to use. (Default: systems git configuration)
            email:
                The git email to use. (Default: systems git configuration)

        Raises:
            RuntimeError when no differences to the previous commit can be found.

        """
        # make a new branch and checkout
        if push_option is None:
            push_option = []

        new_head = create_new_head(repo, DeployManager.retrieve_head_name(active_solution))
        new_head.checkout()

        commit_msg = "Adding new/updated %s" % DeployManager.retrieve_head_name(active_solution)

        add_files_commit_and_push(
            new_head,
            file_paths,
            commit_msg,
            push=not dry_run,
            push_option_list=push_option,
            email=email,
            username=username
        )

    def _get_absolute_zip_path(self, catalog_local_src: str, active_solution: ISolution):
        """ Gets the absolute path to the zip."""
        return Path(catalog_local_src).joinpath(
            self.album.solutions().get_solution_zip_suffix(active_solution.coordinates())
        )

    def _get_absolute_prefix_path(self, catalog: ICatalog, catalog_local_src: str, active_solution: ISolution):
        return Path(catalog_local_src).joinpath(
            self.album.solutions().get_solution_path(catalog, active_solution.coordinates())
        )

    @staticmethod
    def _create_yaml_file_in_local_src(active_solution: ISolution, solution_home: Path):
        """Creates a yaml file in the given repo for the given solution.

        Returns:
            The Path to the created markdown file.

        """
        coordinates = active_solution.coordinates()

        yaml_path = solution_home.joinpath("%s%s" % (coordinates.name(), ".yml"))

        module_logger().debug('Writing yaml file to: %s...' % yaml_path)
        write_dict_to_yml(yaml_path, get_deploy_dict(active_solution))

        return yaml_path

    def _copy_and_zip(self, catalog_local_src: str, active_solution: Solution, folder_path) -> Path:
        """Copies the deploy-file or -folder to the catalog repository.

        Returns:
            The path to the zip.

        """
        zip_path = self._get_absolute_zip_path(catalog_local_src, active_solution)
        module_logger().debug('Creating zip file in: %s...' % str(zip_path))

        if folder_path.is_dir():
            zip_file = Path(zip_folder(folder_path, zip_path))
            return zip_file
        if folder_path.is_file():
            with tempfile.TemporaryDirectory(dir=self._get_tmp_dir()) as tmp_dir:
                tmp_solution_file = Path(tmp_dir).joinpath(DefaultValues.solution_default_name.value)

                copy(folder_path, tmp_solution_file)
                zip_path = zip_paths([tmp_solution_file], zip_path, tmp_dir=self._get_tmp_dir())

            return Path(zip_path)

    def _get_tmp_dir(self):
        return self.album.configuration().cache_path_tmp_internal()

    @staticmethod
    def _copy_files_from_solution(
            active_solution: ISolution, source_path: Path, target_path: Path, name, attribute_path
    ):
        files = []
        source_path = Path(source_path)
        target_path = Path(target_path)
        module_logger().debug('Looking up %s file(s) to copy to %s...' % (name, str(target_path)))

        # get files in solution setup dictionary having the given attribute_path
        file_names = get_dict_entries_from_attribute_path(active_solution.setup(), attribute_path)
        module_logger().debug('Lookup result: %s' % ', '.join(file_names))

        for file_name in file_names:

            file_source_path = source_path.joinpath(file_name)  # relative paths only
            if file_source_path.exists():
                files.append(copy(file_source_path, target_path.joinpath(file_name)))
            else:
                module_logger().error(  # must be an error, otherwise yml of solutions not in sync with folder content
                    'Cannot find %s %s, proceeding without copying...' % (name, file_source_path.absolute()))
        return files

    def _attach_exports(self, catalog: ICatalog, catalog_local_src: str, active_solution: ISolution, deploy_path: Path):
        coordinates = active_solution.coordinates()

        catalog_solution_local_src_path = Path(catalog_local_src).joinpath(
            self.album.configuration().get_solution_path_suffix(coordinates)
        )

        res = []
        res.extend(DeployManager._copy_files_from_solution(
            active_solution, deploy_path, catalog_solution_local_src_path, 'cover', 'covers.source')
        )
        res.extend(DeployManager._copy_files_from_solution(
            active_solution, deploy_path, catalog_solution_local_src_path, 'documentation', 'documentation')
        )
        res.append(create_docker_file(active_solution, catalog_solution_local_src_path))
        res.append(DeployManager._create_yaml_file_in_local_src(active_solution, catalog_solution_local_src_path))
        res.append(create_changelog_file(active_solution, catalog, catalog_solution_local_src_path))

        return res



import os
import re
from contextlib import contextmanager
from pathlib import Path
from typing import Optional, Tuple, List, Generator

import validators
from git import Repo

from album.core.api.model.catalog import ICatalog
from album.core.api.model.catalog_index import ICatalogIndex
from album.core.model.catalog_index import CatalogIndex
from album.core.model.default_values import DefaultValues
from album.core.utils.operations.file_operations import copy, copy_folder, write_dict_to_json, force_remove
from album.core.utils.operations.git_operations import download_repository, init_repository
from album.core.utils.operations.resolve_operations import dict_to_coordinates
from album.core.utils.operations.solution_operations import get_deploy_dict
from album.core.utils.operations.url_operations import download_resource
from album.runner import album_logging
from album.runner.core.api.model.coordinates import ICoordinates
from album.runner.core.api.model.solution import ISolution
from album.runner.core.model.solution import Solution

module_logger = album_logging.get_active_logger


def get_index_url(src, branch_name="main") -> Tuple[str, str]:
    index_src = re.sub(r"\.git$", "", src) + "/-/raw/%s/%s" % (branch_name, DefaultValues.catalog_index_file_name.value)
    index_meta_src = re.sub(r"\.git$", "", src) + "/-/raw/%s/%s" % (
        branch_name, DefaultValues.catalog_index_metafile_json.value
    )
    return index_src, index_meta_src


def get_index_dir(src) -> Tuple[Path, Path]:
    index_src = Path(src).joinpath(DefaultValues.catalog_index_file_name.value)
    index_meta_src = Path(src).joinpath(DefaultValues.catalog_index_metafile_json.value)
    return index_src, index_meta_src


def get_solution_src(src, coordinates: ICoordinates, branch_name="main") -> str:
    """Gets the download link for a solution."""
    return re.sub(r"\.git$", "", src) + "/-/raw/%s/solutions/%s/%s/%s/%s" % (
        branch_name, coordinates.group(), coordinates.name(), coordinates.version(),
        "%s_%s_%s%s" % (coordinates.group(), coordinates.name(), coordinates.version(), ".zip")
    )


class Catalog(ICatalog):

    def __init__(self, catalog_id, name, path, src=None, deletable=True, branch_name="main"):
        """Init routine.

        Args:
            catalog_id:
                The id of the catalog in the collection.
            name:
                The name of the catalog.
            path:
                The absolute path to the catalog.
            src:
                The source of the catalog (Default: None)
                Can be a URL, a git-link, a path to a network-device or a path to any other storage system.
            deletable:
                Boolean to indicate whether the catalog is deletable or not. Relevant for a collection of catalogs.
            branch_name:
                When a git based catalog this attribute can be set to use other branches than the main branch (default)
        """
        self._catalog_id = catalog_id
        self._name = name
        self._src = src
        self._version = None  # is set automatically with the index
        self._catalog_index: Optional[ICatalogIndex] = None
        self._path = Path(path)
        self._branch_name = branch_name

        self._is_deletable = deletable

        self._solution_list_path = self._path.joinpath(DefaultValues.catalog_solution_list_file_name.value)
        self._meta_path = self._path.joinpath(DefaultValues.catalog_index_metafile_json.value)
        self._index_path = self._path.joinpath(DefaultValues.catalog_index_file_name.value)

        if self.is_local() and self._src:
            self._src = Path(self._src).absolute()

    def __eq__(self, other):
        return isinstance(other, ICatalog) and \
               other.catalog_id() == self._catalog_id

    def __del__(self):
        self.dispose()

    def dispose(self):
        if self._catalog_index is not None:
            self._catalog_index.close()

    def is_cache(self):
        """Returns Boolean indicating whether the catalog is used for caching only."""
        return self._src is None or self.is_local() and self._path.exists() and os.path.samefile(
            str(self._src), self._path
        )

    def is_local(self):
        """Returns Boolean indicating whether the catalog is remote or local."""
        return not validators.url(str(self._src))

    def update_index_cache_if_possible(self):
        try:
            self.update_index_cache()
        except AssertionError:
            module_logger().warning("Could not refresh index. Source invalid!")
            return False
        except ConnectionError:
            module_logger().warning("Could not refresh index. Connection error!")
            return False
        except FileNotFoundError:
            module_logger().warning("Could not refresh index. Source not found!")
            return False
        except Exception as e:
            module_logger().warning("Could not refresh index. Unknown reason!")
            module_logger().warning(e)
            return False

        return True

    def update_index_cache(self):
        if self.is_cache():
            return False

        if self.is_local():
            index_available = self.copy_index_from_src_to_cache()
        else:
            index_available = self.download_index()

        if not index_available:
            self.dispose()
            # index got deleted in src so we do the same locally
            force_remove(self._index_path)
        return True

    def add(self, active_solution: ISolution, force_overwrite=False):
        solution_attrs = get_deploy_dict(active_solution)

        if active_solution.setup().doi:
            solution_attrs["doi"] = active_solution.setup().doi
            if self._catalog_index.get_solution_by_doi(solution_attrs["doi"]) is not None:
                if force_overwrite:
                    module_logger().warning("Solution already exists! Overwriting...")
                else:
                    raise RuntimeError("DOI already exists in catalog! Aborting...")

        if active_solution.setup().deposit_id:
            solution_attrs["deposit_id"] = active_solution.setup().deposit_id

        lookup_solution = self._catalog_index.get_solution_by_coordinates(active_solution.coordinates())
        if lookup_solution:
            if force_overwrite:
                module_logger().warning("Solution already exists! Overwriting...")
            else:
                raise RuntimeError("Solution already exists in catalog! Aborting...")

        self._catalog_index.update(active_solution.coordinates(), solution_attrs)
        self._catalog_index.save()
        self._catalog_index.export(self._solution_list_path)

    def remove(self, active_solution: ISolution):
        if self.is_local():
            solution_attrs = get_deploy_dict(active_solution)
            solution_entry = self._catalog_index.remove_solution_by_group_name_version(
                dict_to_coordinates(solution_attrs)
            )
            if solution_entry:
                self._catalog_index.export(self._solution_list_path)
            else:
                module_logger().warning("Solution not installed! Doing nothing...")
        else:
            module_logger().warning("Cannot remove entries from a remote catalog! Doing nothing...")

    def copy_index_from_src_to_cache(self) -> bool:
        src_path_index = Path(self._src).joinpath(DefaultValues.catalog_index_file_name.value)
        src_path_meta = Path(self._src).joinpath(DefaultValues.catalog_index_metafile_json.value)

        # check if meta information valid, only then continue
        if not src_path_meta.exists():
            raise FileNotFoundError("Could not find file %s..." % src_path_meta)

        module_logger().debug("Copying meta information from %s to %s..." % (src_path_meta, self._meta_path))
        copy(src_path_meta, self._meta_path)

        if src_path_index.exists():
            module_logger().debug("Copying index from %s to %s..." % (src_path_index, self._index_path))
            copy(src_path_index, self._index_path)
        else:
            if self._index_path.exists():
                # case index was deleted in the src
                return False
            else:
                module_logger().debug("Index file of the catalog does not exist yet...")
        return True

    def copy_index_from_cache_to_src(self):
        src_path_index = Path(self._src).joinpath(DefaultValues.catalog_index_file_name.value)

        if not src_path_index.exists():
            if not self._index_path.parent.exists():
                self._index_path.parent.mkdir(parents=True)

        module_logger().debug("Copying index from %s to %s..." % (self._index_path, src_path_index))
        copy(self._index_path, src_path_index)
        src_path_solution_list = Path(self._src).joinpath(DefaultValues.catalog_solution_list_file_name.value)

        module_logger().debug(
            "Copying exported index from %s to %s..." % (self._solution_list_path, src_path_solution_list)
        )
        copy(self._solution_list_path, src_path_solution_list)

    def download_index(self) -> bool:
        src, meta_src = get_index_url(self._src, self._branch_name)

        # check metadata first before moving on
        download_resource(meta_src, self._meta_path)

        try:
            download_resource(src, self._index_path)
        except AssertionError:
            # catalogs don't necessary have an index file. There simply might not be one yet or it got deleted.
            return False
        return True

    @contextmanager
    def retrieve_catalog(self, path=None, force_retrieve=False, update=True) -> Generator[Repo, None, None]:
        if self.is_cache():
            raise RuntimeError("Cannot retrieve a cache catalog as no source exists!")

        path = Path(path) if path else self._path

        if self.is_local():  # case src is not downloadable
            copy_folder(self._src, path, copy_root_folder=False, force_copy=True)
            repo = init_repository(path)
        else:  # case src is downloadable
            module_logger().debug("Trying to retrieve catalog %s to the path %s..." % (self._name, str(path)))
            repo = download_repository(self._src, str(path), force_download=force_retrieve, update=update)

        yield repo
        repo.close()

    def write_catalog_meta_information(self):
        d = self.get_meta_information()
        write_dict_to_json(self._meta_path, d)

    def get_meta_information(self):
        return {
            "name": self._name,
            "version": self._version
        }

    def get_all_solution_versions(self, group: str, name: str) -> List[Solution]:
        versions = self._catalog_index.get_all_solution_versions(group, name)
        res = []
        for version in versions:
            res.append(Solution(attrs=version))
        return res

    def load_index(self):
        self._catalog_index = CatalogIndex(self._name, self._index_path)

    def catalog_id(self) -> int:
        return self._catalog_id

    def name(self) -> str:
        return self._name

    def src(self) -> Path:
        return self._src

    def version(self) -> str:
        return self._version

    def index(self) -> ICatalogIndex:
        return self._catalog_index

    def path(self) -> Path:
        return self._path

    def branch_name(self) -> str:
        return self._branch_name

    def is_deletable(self) -> bool:
        return self._is_deletable

    def solution_list_path(self) -> Path:
        return self._solution_list_path

    def index_path(self) -> Path:
        return self._index_path

    def set_catalog_id(self, catalog_id):
        self._catalog_id = catalog_id

    def set_version(self, version):
        self._version = version



import pathlib
from typing import Optional, Union

from album.core.api.model.catalog import ICatalog
from album.core.api.model.collection_index import ICollectionIndex
from album.core.api.model.collection_solution import ICollectionSolution
from album.core.api.model.configuration import IConfiguration
from album.core.controller.album_controller import AlbumController
from album.core.utils.core_logging import configure_root_logger
from album.runner.album_logging import pop_active_logger, LogLevel
from album.runner.core.api.model.script_creator import IScriptCreator
from album.runner.core.api.model.solution import ISolution


class Album:
    class Builder:
        _base_cache_path: Optional[Union[str, pathlib.Path]] = None
        _log_format: str = None
        _log_format_time: str = None
        _log_level: LogLevel = None

        def base_cache_path(self, base_cache_path: Union[str, pathlib.Path]) -> 'Album.Builder':
            self._base_cache_path = base_cache_path
            return self

        def log_format(self, log_format: str) -> 'Album.Builder':
            self._log_format = log_format
            return self

        def log_format_time(self, log_format_time: str) -> 'Album.Builder':
            self._log_format_time = log_format_time
            return self

        def log_level(self, log_level: LogLevel) -> 'Album.Builder':
            self._log_level = log_level
            return self

        def build(self) -> 'Album':
            return Album(self)

    def __init__(self, builder: 'Album.Builder') -> None:
        self._options = builder
        self._controller = AlbumController(self._options._base_cache_path)
        self.logger_pushed = True
        configure_root_logger(log_format=self._options._log_format, log_format_time=self._options._log_format_time,
                              log_level=self._options._log_level)

    def resolve(self, resolve_solution: str) -> ICollectionSolution:
        return self._controller.collection_manager().resolve_and_load(resolve_solution)

    def resolve_installed(self, resolve_solution: str) -> ICollectionSolution:
        return self._controller.collection_manager().resolve_installed_and_load(resolve_solution)

    def load_or_create_collection(self):
        self._controller.collection_manager().load_or_create()

    def get_index_as_dict(self):
        return self._controller.collection_manager().get_index_as_dict()

    def get_catalogs_as_dict(self):
        return self._controller.collection_manager().catalogs().get_all_as_dict()

    def get_catalog_by_name(self, name) -> ICatalog:
        return self._controller.catalogs().get_by_name(name)

    def get_catalog_by_src(self, src) -> ICatalog:
        return self._controller.catalogs().get_by_src(src)

    def get_collection_index(self) -> ICollectionIndex:
        return self._controller.collection_manager().get_collection_index()

    def test(self, collection_solution: ICollectionSolution, args=None):
        return self._controller.test_manager().test(collection_solution, args)

    def load(self, path) -> Optional[ISolution]:
        return self._controller.state_manager().load(path)

    def search(self, keywords):
        """Searches through album catalogs to find closest matching solution.
        """
        return self._controller.search_manager().search(keywords)

    def run(self, collection_solution: ICollectionSolution, argv=None, run_immediately=False):
        return self._controller.run_manager().run(collection_solution, run_immediately, argv)

    def install(self, collection_solution: ICollectionSolution, argv=None):
        return self._controller.install_manager().install(collection_solution, argv)

    def uninstall(self, collection_solution: ICollectionSolution, rm_dep=False, argv=None):
        """Removes a solution from the disk. Thereby uninstalling its environment and deleting all its downloads.

        Args:
            argv:
                Arguments which should be appended to the script call
            collection_solution:
                The solution to remove
            rm_dep:
                Boolean to indicate whether to remove parents too.

        """
        return self._controller.install_manager().uninstall(collection_solution, rm_dep, argv)

    def deploy(self, deploy_path: str, catalog_name: str, dry_run: bool, push_option=None, git_email: str = None,
               git_name: str = None, force_deploy: bool = False, changelog: str = ""):
        """Function corresponding to the `deploy` subcommand of `album`.

        Generates the yml for a album and creates a merge request to the catalog only
        including the yaml and solution file.

        Args:
            force_deploy:
                Force overwrites a existing solution during deployment. Only for local catalogs.
            deploy_path:
                Path to a directory or a file.
                If directory: Must contain "solution.py" file.
            catalog_name:
                The catalog to deploy to. Either specify via argument in deploy-call, via url in solution or use
                default catalog.
            dry_run:
                When set, prepares deployment in local src of the catlog (creating zip, docker, yml),
                but not adding to the catalog src.
            push_option:
                Push options for the catalog repository.
            git_email:
                The git email to use. (Default: systems git configuration)
            git_name:
                The git user to use. (Default: systems git configuration)
            changelog:
                The change associated with this version of a solution compared to the last version.

        """
        return self._controller.deploy_manager().deploy(deploy_path=deploy_path, catalog_name=catalog_name, dry_run=dry_run,
                                                        push_options=push_option, git_email=git_email, git_name=git_name,
                                                        force_deploy=force_deploy, changelog=changelog)

    def clone(self, path: str, target_dir: str, name: str) -> None:
        """
        Function corresponding to the `clone` subcommand of `album`.

        Args:
            path: the source of the clone command - a solution (group:name:version, path, or URL to file) or a catalog
                template string (i.e. template:catalog)
            target_dir: the directory where the cloned solution or catalog will be added to
            name: the name of the solution or catalog to be created

        """
        return self._controller.clone_manager().clone(path, target_dir, name)

    def close(self):
        if self.logger_pushed:
            pop_active_logger()
        self._controller.close()

    def run_solution_script(self, resolve_result: ICollectionSolution, script: IScriptCreator):
        self._controller.script_manager().run_solution_script(resolve_result, script)

    def upgrade(self, catalog_name=None, dry_run=False):
        return self._controller.collection_manager().catalogs().update_collection(catalog_name, dry_run)

    def update(self, catalog_name=None):
        return self._controller.collection_manager().catalogs().update_any(catalog_name)

    def add_catalog(self, catalog_src):
        return self._controller.collection_manager().catalogs().add_by_src(catalog_src)

    def remove_catalog_by_src(self, catalog_src):
        return self._controller.collection_manager().catalogs().remove_from_collection_by_src(catalog_src)

    def remove_catalog_by_name(self, catalog_src):
        return self._controller.collection_manager().catalogs().remove_from_collection_by_name(catalog_src)

    def configuration(self) -> IConfiguration:
        return self._controller.configuration()

    def is_installed(self, resolve_result: ICollectionSolution):
        return self._controller.solutions().is_installed(resolve_result.catalog(), resolve_result.coordinates())

    def load_catalog_index(self, catalog: ICatalog):
        return self._controller.migration_manager().load_index(catalog)



import logging
from typing import Dict

import colorlog

from album.core.api.model.catalog_updates import ICatalogUpdates
from album.runner.core.api.model.solution import ISolution


def get_solution_as_string(solution: ISolution, solution_path):
    param_example_str = ''
    setup = solution.setup()
    if setup.args:
        for arg in setup.args:
            param_example_str += '--%s PARAMETER_VALUE ' % arg['name']
    res = 'Solution details about %s:\n\n' % solution_path
    if setup.title:
        res += '%s\n' % setup.title
        res += '%s\n' % ('=' * len(setup.title))
    if setup.description:
        res += '%s\n\n' % setup.description
    res += 'Group            : %s\n' % setup.group
    res += 'Name             : %s\n' % setup.name
    res += 'Version          : %s' % setup.version
    res += '%s' % get_credit_as_string(solution)
    res += 'Solution metadata:\n\n'
    if setup.authors:
        res += 'Solution authors : %s\n' % ", ".join(setup.authors)
    if setup.license:
        res += 'License          : %s\n' % setup.license
    if setup.acknowledgement:
        res += 'Acknowledgement  : %s\n' % setup.acknowledgement
    if setup.tags:
        res += 'Tags             : %s\n' % ", ".join(setup.tags)
    res += '\n'
    res += 'Usage:\n\n'
    res += '  album install %s\n' % solution_path
    res += '  album run %s %s\n' % (solution.coordinates(), param_example_str)
    res += '  album test %s\n' % solution.coordinates()
    res += '  album uninstall %s\n' % solution.coordinates()
    res += '\n'
    if setup.args:
        res += 'Run parameters:\n\n'
        for arg in setup.args:
            res += '  --%s: %s\n' % (arg["name"], arg["description"])
    return res


def get_credit_as_string(solution: ISolution) -> str:
    res = ''
    if solution.setup().cite:
        res += '\n\nCredits:\n\n'
        for citation in solution.setup().cite:
            text = citation['text']
            if 'doi' in citation:
                text += ' (DOI: %s)' % citation['doi']
            res += '%s\n' % text
        res += '\n'
    return res


def get_updates_as_string(updates: Dict[str, ICatalogUpdates]) -> str:
    res = ''
    for catalog_name in updates:
        change = updates[catalog_name]
        res += 'Catalog: %s\n' % change.catalog().name()
        if len(change.catalog_attribute_changes()) > 0:
            res += '  Catalog attribute changes:\n'
            for item in change.catalog_attribute_changes():
                res += '  name: %s, new value: %s\n' % (item.attribute(), item.new_value())
        if len(change.solution_changes()) > 0:
            res += '  Catalog solution changes:\n'
            for i, item in enumerate(change.solution_changes()):
                if i is len(change.solution_changes()) - 1:
                    res += '  └─ [%s] %s\n' % (item.change_type().name, item.coordinates())
                    separator = ' '
                else:
                    res += '  ├─ [%s] %s\n' % (item.change_type().name, item.coordinates())
                    separator = '|'
                res += '  %s     %schangelog: %s\n' % (
                    separator, (" " * len(item.change_type().name)), item.change_log())

        if len(change.catalog_attribute_changes()) == 0 and len(change.solution_changes()) == 0:
            res += '  No changes.\n'
    return res


def get_index_as_string(index_dict: dict):
    res = '\n'
    if 'catalogs' in index_dict:
        for catalog in index_dict['catalogs']:
            res += 'Catalog \'%s\':\n' % catalog['name']
            res += '├─ name: %s\n' % catalog['name']
            res += '├─ src: %s\n' % catalog['src']
            res += '├─ catalog_id: %s\n' % catalog['catalog_id']
            if len(catalog['solutions']) > 0:
                res += '├─ deletable: %s\n' % catalog['deletable']
                res += '└─ [installed] solutions:\n'
                for i, solution in enumerate(catalog['solutions']):
                    installed = ' '
                    if solution['internal']['installed']:
                        installed = 'x'
                    if i is len(catalog['solutions']) - 1:
                        res += '   └─ [%s] %s:%s:%s\n' % (
                            installed, solution['setup']['group'], solution['setup']['name'], solution['setup']['version'])
                    else:
                        res += '   ├─ [%s] %s:%s:%s\n' % (
                            installed, solution['setup']['group'], solution['setup']['name'], solution['setup']['version'])
            else:
                res += '└─ deletable: %s\n' % catalog['deletable']
    return res


def get_search_result_as_string(args, search_result):
    res = ''
    if len(search_result) > 0:
        res += 'Search results for "%s" - run `album info SOLUTION_ID` for more information:\n' % ' '.join(
            args.keywords)
        res += '[SCORE] SOLUTION_ID\n'
        for result in search_result:
            res += '[%s] %s\n' % (result[1], result[0])
    else:
        res += 'No search results for "%s".' % ' '.join(args.keywords)
    return res


def get_logging_formatter(fmt=None, time=None):
    if not fmt:
        fmt = '%(log_color)s%(asctime)s %(levelname)-7s %(shortened_name)s%(message)s'
    if not time:
        time = '%H:%M:%S'
    return colorlog.ColoredFormatter(fmt, time, log_colors={
        'DEBUG': 'cyan',
        'WARNING': 'yellow',
        'ERROR': 'red',
        'CRITICAL': 'red,bold',
    })


def get_logger_name_minimizer_filter():
    class NameDotFilter(logging.Filter):
        def filter(self, record):
            count = record.name.count('.') + record.name.count('~')
            if count > 0:
                record.shortened_name = '~' * count + ' '
            else:
                record.shortened_name = ''
            return True

    return NameDotFilter()


def get_message_filter():
    class MessageFilter(logging.Filter):
        def filter(self, record):
            self._apply_mamba_menuinst_filter(record)
            return True

        def _apply_mamba_menuinst_filter(self, record):
            if record.msg and isinstance(record.msg, str) and 'menuinst called from non-root env' in record.msg:
                record.levelname = 'WARNING'
                record.levelno = logging.getLevelName('WARNING')

    return MessageFilter()



import argparse
from queue import Queue, Empty
from typing import List

from album.core.api.controller.controller import IAlbumController
from album.core.api.controller.script_manager import IScriptManager
from album.core.api.model.collection_solution import ICollectionSolution
from album.core.controller.run_manager import SolutionGroup
from album.core.model.script_queue_entry import ScriptQueueEntry
from album.core.utils.operations.resolve_operations import build_resolve_string
from album.core.utils.operations.solution_operations import get_steps_dict, get_parent_dict
from album.runner.album_logging import get_active_logger
from album.runner.core.api.model.script_creator import IScriptCreator
from album.runner.core.api.model.solution import ISolution
from album.runner.core.model.script_creator import ScriptCreatorRun, ScriptCreatorRunWithParent

module_logger = get_active_logger


class ScriptManager(IScriptManager):

    def __init__(self, album: IAlbumController):
        self.album = album

    def run_solution_script(self, resolve_result: ICollectionSolution, script_creator: ScriptCreatorRun):
        queue = Queue()
        self.build_queue(resolve_result, queue, script_creator, False, [""])
        script_queue_entry = queue.get(block=False)
        self.album.environment_manager().run_scripts(script_queue_entry.environment, script_queue_entry.scripts,
                                                     pipe_output=False)

    def run_queue(self, queue: Queue):
        module_logger().debug("Running queue...")
        try:
            while True:
                script_queue_entry = queue.get(block=False)
                module_logger().debug("Running task \"%s\"..." % script_queue_entry.coordinates.name())
                self._run_in_environment(script_queue_entry)
                module_logger().debug("Finished running task \"%s\"!" % script_queue_entry.coordinates.name())
                queue.task_done()
        except Empty:
            module_logger().debug("Currently nothing more to run!")

    def build_queue(self, collection_solution: ICollectionSolution, queue, script_creator: ScriptCreatorRun,
                    run_immediately=False,
                    argv=None):
        if argv is None:
            argv = [""]
        solution = collection_solution.loaded_solution()
        steps = get_steps_dict(solution)
        if steps:  # solution consists of at least one step
            # a step base album is first initialized in the album environment to be able to harvest it's arguments
            # active_solution.init() THIS FEATURE IS TEMPORARY DISABLED

            step_solution_parsed_args = self.__parse_args(solution, argv)
            module_logger().debug("Building queue for %s steps.." % len(steps))

            for i, step in enumerate(steps):
                module_logger().debug("Adding step %s / %s to queue..." % (i, len(steps)))
                if type(step) is list:
                    self._build_steps_queue(queue, step, script_creator, run_immediately, step_solution_parsed_args)
                else:
                    self._build_steps_queue(queue, [step], script_creator, run_immediately, step_solution_parsed_args)
        else:  # single element queue, no steps
            parent = get_parent_dict(solution)
            if parent:
                module_logger().debug("Adding standalone solution with parent to queue...")
                # create script with parent
                queue.put(
                    self._create_solution_run_with_parent_script_standalone(collection_solution, argv, script_creator))
            else:
                module_logger().debug("Adding standalone to queue...")
                # create script without parent
                queue.put(self._create_solution_run_script_standalone(collection_solution, argv, script_creator))

    def _build_steps_queue(
            self, queue: Queue, steps: list, script_creator: ScriptCreatorRun, run_immediately=False,
            step_solution_parsed_args=None
    ):
        """Builds the queue of step-album to be executed. FIFO queue expected.

        Args:
            script_creator:
                The ScriptCreatorRun object to use to create the execution script.
            step_solution_parsed_args:
                Namespace object from parsing the step-solution arguments
            queue:
                The queue object.
            steps:
                The steps of a stepped solution.
            run_immediately:
                Boolean. When true, a collection is run immediately without solving for further steps and pushing them
                to the queue. Can result in resolving problems in further downstream collections. Necessary for
                branching decisions.

        """
        # start with an empty collection of steps with the same parent
        same_parent_step_collection = SolutionGroup(step_solution_parsed_args)
        citations = []

        for step in steps:
            module_logger().debug('resolving step \"%s\"...' % step["name"])
            resolve_result = self.album.collection_manager().resolve_installed_and_load(
                build_resolve_string(step))

            citations.append(resolve_result.loaded_solution())

            parent = get_parent_dict(resolve_result.loaded_solution())
            if parent:  # collect steps as long as they have the same parent
                current_parent_script_resolve = self.album.collection_manager().resolve_installed_and_load(
                    build_resolve_string(parent)
                )

                # check whether the step has the same parent as the previous steps
                if same_parent_step_collection.parent and \
                        same_parent_step_collection.parent.coordinates() != current_parent_script_resolve.coordinates():

                    # put old collection to queue
                    queue.put(self._create_solution_run_collection_script(same_parent_step_collection, script_creator))

                    # runs the old collection immediately
                    if run_immediately:
                        self.run_queue(queue)

                    # set new parent
                    same_parent_step_collection.parent = current_parent_script_resolve

                    # overwrite old steps
                    same_parent_step_collection.steps_solution = [resolve_result.loaded_solution()]
                    same_parent_step_collection.steps = [step]

                else:  # same or new parent
                    module_logger().debug('Pushed step \"%s\" in queue...' % step["name"])

                    # set parent
                    same_parent_step_collection.parent = current_parent_script_resolve

                    # append another step to the steps already having the same parent
                    same_parent_step_collection.steps_solution.append(resolve_result.loaded_solution())
                    same_parent_step_collection.steps.append(step)

            else:  # add a step without collection (also parent)
                # put collection (if any) to queue
                if same_parent_step_collection.parent:
                    queue.put(self._create_solution_run_collection_script(same_parent_step_collection, script_creator))

                # empty the collection for possible next steps
                same_parent_step_collection = SolutionGroup(step_solution_parsed_args)

                # run the old collection immediately
                if run_immediately:
                    self.run_queue(queue)

                # harvest arguments in the description of the step
                step_args = self._get_args(step, step_solution_parsed_args[0])

                # add step without parent
                queue.put(
                    self._create_solution_run_script_standalone(resolve_result, step_args, script_creator)
                )

        # put rest to queue
        if same_parent_step_collection.parent:
            queue.put(self._create_solution_run_collection_script(same_parent_step_collection, script_creator))

        self._print_credit(citations)

        # run the old collection immediately
        if run_immediately:
            self.run_queue(queue)

    def _create_solution_run_script_standalone(
            self, collection_solution: ICollectionSolution, args: list, script_creator: IScriptCreator
    ) -> ScriptQueueEntry:
        """Creates the execution script for a album object giving its arguments.

        Args:
            script_creator:
                The ScriptCreator object to use to create the execution script.
            collection_solution:
                The collection solution object to create the executable script for.
            args:
                The arguments integrated in the script.

        Returns:
                The solution object and its scripts (in a list)

        """
        module_logger().debug('Creating standalone album script \"%s\"...' % collection_solution.coordinates().name())

        self._print_credit([collection_solution.loaded_solution()])
        environment = self.album.environment_manager().set_environment(collection_solution)
        script = script_creator.create_script(collection_solution.loaded_solution(), args)

        return ScriptQueueEntry(collection_solution.coordinates(), [script], environment=environment)

    def _create_solution_run_with_parent_script_standalone(
            self, collection_solution: ICollectionSolution, args: list, script_creator: ScriptCreatorRun
    ) -> ScriptQueueEntry:
        """Creates the execution script for a album object having a parent dependency giving its arguments.

        Args:
            script_creator:
                The ScriptCreatorRun object to use to create the execution script.
            collection_solution:
                The collection solution object to create the executable script for.
            args:
                The arguments integrated in the script.

        Returns:
                The solution object and its scripts (in a list).

        """
        environment = self.album.environment_manager().set_environment(collection_solution)
        active_solution = collection_solution.loaded_solution()

        # create script
        script = script_creator.create_script(collection_solution.loaded_solution(), args)

        # TODO this should probably move into the runner
        self._print_credit([active_solution])

        return ScriptQueueEntry(collection_solution.loaded_solution().coordinates(), [script], environment)

    def _create_solution_run_collection_script(
            self, solution_collection: SolutionGroup, script_creator: ScriptCreatorRun
    ) -> ScriptQueueEntry:
        """Creates the execution script for a collection of solutions all having the same parent dependency.

        Args:
            script_creator:
                The ScriptCreator object to use to create the execution script.
            solution_collection
        Returns:
            The solution shared parent object and its scripts.

        """
        # load parent & steps
        parent_solution = self.album.state_manager().load(solution_collection.parent.path())
        self.album.solutions().set_cache_paths(parent_solution,
                                                            solution_collection.parent.catalog())

        environment = self.album.environment_manager().set_environment(solution_collection.parent)
        module_logger().debug('Creating script for steps (%s) with parent \"%s\"...' % (
            ", ".join([s.coordinates().name() for s in solution_collection.steps_solution]),
            parent_solution.coordinates().name()))

        # handle arguments
        parsed_parent_args, parsed_steps_args_list = self._resolve_args(
            parent_solution=parent_solution,
            steps_solution=solution_collection.steps_solution,
            steps=solution_collection.steps,
            step_solution_parsed_args=solution_collection.parent_parsed_args,
        )

        # create script
        scripts = self._create_solution_run_with_parent_script(
            parent_solution, parsed_parent_args,
            solution_collection.steps_solution,
            parsed_steps_args_list,
            script_creator
        )

        return ScriptQueueEntry(parent_solution.coordinates(), scripts, environment)

    @staticmethod
    def _create_solution_run_with_parent_script(
            parent_solution: ISolution,
            parent_args: list,
            child_solution_list: List[ISolution],
            child_args: list,
            script_creator: ScriptCreatorRun
    ):
        """Creates the script for the parent solution as well as for all its steps (child solutions).

        Args:
            script_creator:
                The ScriptCreatorRun object to use to create the execution script.
            parent_solution:
                The parent solution object.
            parent_args:
                Arguments to use for the parent call.
            child_solution_list:
                A list of all solution objects to be executed with the same parent.
            child_args:
                List of arguments for the call of each child solution.

        Returns:
            A list holding all execution scripts.

        """
        module_logger().debug('Creating album script with parent \"%s\"...' % parent_solution.coordinates().name())

        script_creator_run_with_parent = ScriptCreatorRunWithParent(script_creator, child_solution_list, child_args)
        script = script_creator_run_with_parent.create_script(parent_solution, parent_args)
        script_list = [script]

        return script_list

    def __parse_args(self, active_solution: ISolution, args: list):
        """Parse arguments of loaded solution."""
        parser = argparse.ArgumentParser()

        class FileAction(argparse.Action):
            def __init__(self, option_strings, dest, nargs=None, **kwargs):
                if nargs is not None:
                    raise ValueError("nargs not allowed")
                super(FileAction, self).__init__(option_strings, dest, **kwargs)

            def __call__(self, p, namespace, values, option_string=None):
                setattr(namespace, self.dest, active_solution.get_arg(self.dest)['action'](values))

        for element in active_solution.setup()["args"]:
            if 'action' in element.keys():
                parser.add_argument("--" + element["name"], action=FileAction)
            else:
                parser.add_argument("--" + element["name"])

        return parser.parse_known_args(args=args)

    def _resolve_args(
            self,
            parent_solution: ISolution,
            steps_solution: List[ISolution],
            steps: list,
            step_solution_parsed_args: list,
            args=None
    ):
        """Resolves arguments of all steps and their parents."""
        args = [] if args is None else args
        parsed_parent_args = None
        parsed_steps_args_list = []

        module_logger().debug('Parsing arguments...')

        # iterate over all steps and parse arguments together
        for idx, step_solution in enumerate(steps_solution):
            step_parser = argparse.ArgumentParser()

            # the steps description
            step = steps[idx]

            if step:  # case steps argument resolving
                step_args = self._get_args(step, step_solution_parsed_args[0])
            else:  # case single step solution
                step_args = args

            # add parent arguments to the step album object arguments
            parent_dict = get_parent_dict(step_solution)
            if 'args' in parent_dict:
                for param in parent_dict["args"]:
                    step_args.insert(0, f"--{param['name']}={str(param['value'])}")

            # add parent arguments
            if 'args' in parent_solution.setup():
                [step_parser.add_argument("--" + element["name"]) for element in parent_solution.setup()["args"]]

            # parse all known arguments
            args_known, args_unknown = step_parser.parse_known_args(step_args)

            # only set parents args if not already set
            if not parsed_parent_args:
                parsed_parent_args = [""]
                parsed_parent_args.extend(
                    ["--" + arg_name + "=" + getattr(args_known, arg_name) for arg_name in vars(args_known)])
                module_logger().debug(
                    'For step \"%s\" set parent arguments to %s...' % (
                        step_solution.coordinates().name(), parsed_parent_args
                    )
                )

            # args_unknown are step args
            parsed_steps_args_list.append(args_unknown)
            module_logger().debug(
                'For step \"%s\" set step arguments to %s...' % (step_solution.coordinates().name(), args_unknown))

        return parsed_parent_args, parsed_steps_args_list

    def _run_in_environment(self, script_queue_entry: ScriptQueueEntry):
        """Pushes a new logger to the stack before running the solution and pops it afterwards."""
        module_logger().debug(
            "Running script in environment of solution \"%s\"..." % script_queue_entry.coordinates.name())
        self.album.environment_manager().run_scripts(script_queue_entry.environment, script_queue_entry.scripts)
        module_logger().debug(
            "Done running script in environment of solution \"%s\"..." % script_queue_entry.coordinates.name())

    @staticmethod
    def _get_args(step, args):
        """Parse callable arguments belonging to a step into a list of strings"""
        argv = [""]
        if 'args' in step:
            for param in step["args"]:
                argv.append(f"--{param['name']}={str(param['value'](args))}")
        return argv

    @staticmethod
    def _print_credit(active_solutions: List[ISolution]) -> None:
        res = ScriptManager._get_credit_as_string(active_solutions)
        if res:
            module_logger().info(res)

    @staticmethod
    def _get_credit_as_string(active_solutions: List[ISolution]):
        res = ''
        for active_solution in active_solutions:
            if active_solution.setup().cite:
                for citation in active_solution.setup().cite:
                    text = citation['text']
                    if 'doi' in citation:
                        text += ' (DOI: %s)' % citation['doi']
                    res += '%s\n' % text
        if len(res) > 0:
            return '\n\nSolution credits:\n\n%s\n' % res
        return None



from queue import Queue
from typing import List

from album.core.api.controller.controller import IAlbumController
from album.core.api.controller.run_manager import IRunManager
from album.core.api.model.collection_solution import ICollectionSolution
from album.runner import album_logging
from album.runner.core.api.model.solution import ISolution
from album.runner.core.model.script_creator import ScriptCreatorRun

module_logger = album_logging.get_active_logger


class SolutionGroup:

    def __init__(
            self,
            parent_parsed_args=None,
            parent: ICollectionSolution = None,
            steps_solution=None,
            steps=None
    ):
        if parent_parsed_args is None:
            parent_parsed_args = [None]
        if steps_solution is None:
            steps_solution = []
        if steps is None:
            steps = []
        self.parent_parsed_args = parent_parsed_args
        self.parent = parent
        self.steps_solution: List[ISolution] = steps_solution  # The solution objects of all steps.

        self.steps = steps  # The step description of the step. Must hold the argument keyword.

    def __eq__(self, o: object) -> bool:
        return isinstance(o, SolutionGroup) and \
               o.parent.coordinates() == self.parent.coordinates() and \
               o.steps_solution == self.steps_solution and \
               o.steps == self.steps


class RunManager(IRunManager):

    def __init__(self, album: IAlbumController):
        self.album = album

        self.init_script = ""

    def run(self, resolve_result: ICollectionSolution, run_immediately=False, argv=None):
        """Run an already loaded solution."""
        if not resolve_result.catalog:
            module_logger().debug('solution loaded locally: %s...' % str(resolve_result.loaded_solution().coordinates()))
        else:
            module_logger().debug('solution loaded from catalog: \"%s\"...' % str(resolve_result.loaded_solution().coordinates()))
        module_logger().debug(
            "Initializing script to run \"%s\"" % resolve_result.loaded_solution().coordinates())

        if argv is None:
            argv = [""]

        # pushing album and their scripts to a queue
        que = Queue()

        # builds the queue
        self.album.script_manager().build_queue(resolve_result, que, ScriptCreatorRun(), run_immediately, argv)

        # runs the queue
        self.album.script_manager().run_queue(que)

        module_logger().debug(
            "Finished running script for solution \"%s\"" % resolve_result.loaded_solution().coordinates().name())



from datetime import datetime
from pathlib import Path

import keepachangelog
from album.core.model.default_values import DefaultValues

from album.core.api.model.catalog import ICatalog
from album.core.utils.operations.file_operations import create_path_recursively
from album.runner.album_logging import get_active_logger
from album.runner.core.api.model.solution import ISolution


def get_changelog_file_name():
    return DefaultValues.changelog_default_name.value


def get_changelog_content(active_solution: ISolution, catalog: ICatalog, dummy_content: str = None):
    content = '''# Changelog
All notable changes to this project will be documented in this file.

The format is based on [Keep a Changelog](https://keepachangelog.com/en/1.0.0/),
and this project adheres to [Semantic Versioning](https://semver.org/spec/v2.0.0.html).
'''
    versions = catalog.get_all_solution_versions(active_solution.coordinates().group(),
                                                 active_solution.coordinates().name())
    for version in versions:
        timestamp = version.setup()['timestamp']
        if timestamp:
            timestamp = datetime.strptime(timestamp, '%Y-%m-%dT%H:%M:%S.%f')
            time = timestamp.strftime('%Y-%m-%d')
            if version.setup().version == active_solution.setup().version and dummy_content:
                change = dummy_content
            else:
                if version.setup()['changelog']:
                    change = version.setup()['changelog']
                else:
                    change = ''
            content += '\n## [%s] - %s\n%s\n' % (version.setup()['version'], time, change)
    return content


def create_changelog_file(active_solution: ISolution, catalog: ICatalog, target_folder: Path):
    """Creates a changelog file in the given repo for the given solution.

    Returns:
        The Path to the created markdown file.

    """
    changelog_path = target_folder.joinpath(get_changelog_file_name())
    get_active_logger().debug('Writing changelog file to: %s...' % changelog_path)
    content = get_changelog_content(active_solution, catalog)
    create_path_recursively(changelog_path.parent)
    with open(str(changelog_path), 'w+') as yml_f:
        yml_f.write(content)
    return changelog_path


def process_changelog_file(catalog: ICatalog, active_solution: ISolution, deploy_path: Path):
    changelog_name = get_changelog_file_name()
    changelog_file = Path(deploy_path).joinpath(changelog_name)
    if changelog_file.exists():
        # process existing changelog file
        changelogs = keepachangelog.to_raw_dict(str(changelog_file))
        keyword = str(active_solution.setup().version).lower()
        if keyword in changelogs:
            active_solution.setup().changelog = changelogs[keyword]['raw']
    else:
        # no changelog file found
        if not active_solution.setup().changelog:
            content = get_changelog_content(active_solution, catalog, '- INSERT LIST OF CHANGES')
            get_active_logger().warning(
                'No %s file found.\nWe recommend documenting changes between versions. '
                'You can either\n\t- use the \'--changelog\' parameter of the deploy command\n\t'
                '- or add a file called %s next to the solution file.\nInsert what\'s printed '
                'between the following lines into %s and add your changes to the version you are '
                'about to release before running \'deploy\':\n\n-----------------\n%s'
                '\n-----------------\n' % (
                    changelog_name, changelog_name, changelog_name, content
                )
            )



from pathlib import Path
from typing import Optional, List

from album.core.api.controller.collection.catalog_handler import ICatalogHandler
from album.core.api.controller.collection.collection_manager import ICollectionManager
from album.core.api.controller.collection.solution_handler import ISolutionHandler
from album.core.api.controller.controller import IAlbumController
from album.core.api.model.catalog import ICatalog
from album.core.api.model.collection_index import ICollectionIndex
from album.core.api.model.collection_solution import ICollectionSolution
from album.core.controller.collection.catalog_handler import CatalogHandler
from album.core.controller.collection.solution_handler import SolutionHandler
from album.core.model.collection_index import CollectionIndex
from album.core.model.default_values import DefaultValues
from album.core.model.resolve_result import ResolveResult
from album.core.utils.operations.file_operations import write_dict_to_json
from album.core.utils.operations.resolve_operations import dict_to_coordinates, get_attributes_from_string, check_doi, \
    get_doi_from_input, check_file_or_url, get_parent, build_resolve_string
from album.runner import album_logging
from album.runner.core.api.model.coordinates import ICoordinates

module_logger = album_logging.get_active_logger


class CollectionManager(ICollectionManager):

    def __init__(self, album: IAlbumController):
        self.album = album
        self.solution_handler = SolutionHandler(self.album)
        self.catalog_handler = CatalogHandler(self.album)
        self.catalog_collection: Optional[ICollectionIndex] = None
        self.collection_loaded = False

    def __del__(self):
        self.close()

    def close(self):
        if self.solution_handler is not None:
            del self.solution_handler
            self.solution_handler = None
        if self.catalog_handler is not None:
            del self.catalog_handler
            self.catalog_handler = None
        if self.catalog_collection is not None:
            self.catalog_collection.close()
            self.catalog_collection = None

    def get_collection_index(self) -> ICollectionIndex:
        return self.catalog_collection

    def load_or_create(self):
        if self.collection_loaded:
            module_logger().warning("CollectionManager().load_or_create_collection() should only be called once.")
            return
        self.collection_loaded = True
        collection_meta = self.album.configuration().get_catalog_collection_meta_dict()
        newly_created = not self.album.configuration().get_catalog_collection_path().exists()
        if collection_meta:
            collection_version = collection_meta["catalog_collection_version"]
        else:
            if newly_created:
                collection_version = CollectionIndex.version
                self.write_version_to_yml(
                    self.album.configuration().get_catalog_collection_meta_path(), "my-collection", collection_version
                )
            else:
                raise RuntimeError(
                    "Album collection database file found, but no meta file specifying the database version."
                )
        if self.catalog_collection is not None:
            self.catalog_collection.close()
        self.catalog_collection = CollectionIndex(name=DefaultValues.catalog_collection_name.value,
                                                  path=self.album.configuration().get_catalog_collection_path())
        self.album.migration_manager().migrate_collection_index(
            self.catalog_collection,
            initial_version=collection_version
        )
        if newly_created:
            self.catalog_handler.add_initial_catalogs()

    def catalogs(self) -> ICatalogHandler:
        return self.catalog_handler

    def solutions(self) -> ISolutionHandler:
        return self.solution_handler

    def get_index_as_dict(self):
        catalogs = self.catalog_collection.get_all_catalogs()
        for catalog in catalogs:
            solutions = []
            for solution in self.catalog_collection.get_solutions_by_catalog(catalog["catalog_id"]):
                solutions.append({
                    'setup': solution.setup(),
                    'internal': solution.internal()
                })
            catalog['solutions'] = solutions
        return {
            'catalogs': catalogs
        }

    def resolve_installed(self, resolve_solution) -> ICollectionSolution:
        resolve_result = self._resolve(resolve_solution)

        if not resolve_result.database_entry():
            raise LookupError("Solution not found!")

        if not resolve_result.database_entry().internal()["installed"]:
            raise ValueError("Solution seems not to be installed! Please install solution first!")

        return resolve_result

    def resolve_installed_and_load(self, resolve_solution) -> ICollectionSolution:
        resolve_result = self.resolve_installed(resolve_solution)

        loaded_solution = self.album.state_manager().load(resolve_result.path())
        self.solution_handler.set_cache_paths(loaded_solution, resolve_result.catalog())

        resolve_result.set_loaded_solution(loaded_solution)

        return resolve_result

    def resolve_and_load(self, resolve_solution) -> ICollectionSolution:
        resolve_result = self._resolve(resolve_solution)
        self.retrieve_and_load_resolve_result(resolve_result)
        self.solution_handler.set_cache_paths(resolve_result.loaded_solution(), resolve_result.catalog())

        return resolve_result

    def resolve_and_load_catalog_coordinates(self, catalog: ICatalog,
                                             coordinates: ICoordinates) -> ICollectionSolution:
        collection_entry = self._search_in_specific_catalog(catalog.catalog_id(), coordinates)
        solution_path = self.solution_handler.get_solution_file(catalog, coordinates)

        resolve_result = ResolveResult(
            path=solution_path, catalog=catalog, collection_entry=collection_entry, coordinates=coordinates
        )
        self.retrieve_and_load_resolve_result(resolve_result)

        self.solution_handler.set_cache_paths(resolve_result.loaded_solution(), resolve_result.catalog())

        return resolve_result

    def resolve_and_load_coordinates(self, coordinates: ICoordinates) -> ICollectionSolution:
        collection_entry = self._search_by_coordinates(coordinates)
        catalog = self.album.catalogs().get_by_id(collection_entry.internal()["catalog_id"])

        solution_path = self.solution_handler.get_solution_file(catalog, coordinates)
        resolve_result = ResolveResult(
            path=solution_path, catalog=catalog, collection_entry=collection_entry, coordinates=coordinates
        )
        self.retrieve_and_load_resolve_result(resolve_result)

        self.solution_handler.set_cache_paths(resolve_result.loaded_solution(), resolve_result.catalog())

        return resolve_result

    def resolve(self, resolve_solution) -> ICollectionSolution:
        resolve_result = self._resolve(resolve_solution)

        if not Path(resolve_result.path()).exists():
            self.solution_handler.retrieve_solution(resolve_result.catalog(), resolve_result.coordinates())

        return resolve_result

    def resolve_parent(self, parent_dict: dict) -> ICollectionSolution:
        # resolve the parent mentioned in the current solution to get its metadata
        resolve_parent_info = build_resolve_string(parent_dict)
        parent_resolve_result = self.resolve_installed_and_load(resolve_parent_info)

        # resolve parent of the parent
        parent = get_parent(parent_resolve_result.database_entry())

        # case parent itself has no further parent
        if parent.internal()["collection_id"] == parent_resolve_result.database_entry().internal()["collection_id"]:

            loaded_solution = self.album.state_manager().load(parent_resolve_result.path())
            self.solution_handler.set_cache_paths(loaded_solution, parent_resolve_result.catalog())

            parent_resolve_result.set_loaded_solution(loaded_solution)

        # case parent itself has another parent
        else:
            parent_catalog = self.album.catalogs().get_by_id(parent.internal()["catalog_id"])
            parent_coordinates = dict_to_coordinates(parent.setup())
            path = self.solution_handler.get_solution_file(parent_catalog, parent_coordinates)

            loaded_solution = self.album.state_manager().load(path)
            self.solution_handler.set_cache_paths(loaded_solution, parent_catalog)

            parent_resolve_result = ResolveResult(
                path=path,
                catalog=parent_catalog,
                collection_entry=parent,
                coordinates=parent_coordinates,
                loaded_solution=loaded_solution
            )

        return parent_resolve_result

    def _resolve(self, str_input) -> ICollectionSolution:
        # always first resolve outside any catalog, excluding a DOI which should be first resolved inside a catalog
        path = check_file_or_url(str_input, self.album.configuration().cache_path_download())

        doi = get_doi_from_input(str_input)
        if path:
            # will load the solution behind the path to get meta-information
            solution_entry = self._search_for_local_file(path)

            # a solution loaded this way will always end up in a local catalog
            catalog = self.album.catalogs().get_local_catalog()
        else:
            # search DOI first
            if doi:
                solution_entry = self._search_doi(doi["doi"])

                # either a doi is found in the collection or it will be downloaded and ends up in a local catalog
                if solution_entry:
                    catalog = self.album.catalogs().get_by_id(solution_entry.internal["catalog_id"])
                else:
                    # download DOI
                    path = check_doi(doi["doi"], self.album.configuration().cache_path_tmp_user())

                    catalog = self.album.catalogs().get_local_catalog()
            else:  # case no doi
                solution_entry = self._search(str_input)

                if not solution_entry:
                    raise LookupError("Solution cannot be resolved in any catalog!")

                catalog = self.album.catalogs().get_by_id(solution_entry.internal()["catalog_id"])

                path = self.solution_handler.get_solution_file(catalog, dict_to_coordinates(solution_entry.setup()))

        coordinates = None

        if solution_entry:
            coordinates = dict_to_coordinates(solution_entry.setup())

        resolve = ResolveResult(
            path=path, catalog=catalog, collection_entry=solution_entry, coordinates=coordinates
        )

        return resolve

    def _search_for_local_file(self, path) -> ICollectionIndex.ICollectionSolution:
        active_solution = self.album.state_manager().load(path)

        # check in collection
        solution_entry = self.catalog_collection.get_solution_by_catalog_grp_name_version(
            self.album.catalogs().get_local_catalog().catalog_id(),
            active_solution.coordinates()
        )

        return solution_entry

    def _search_doi(self, doi):
        solution_entry = self.catalog_collection.get_solution_by_doi(doi)

        return solution_entry

    def _search(self, str_input) -> ICollectionIndex.ICollectionSolution:
        """Searches ONLY in the catalog collection, given a string which is interpreted."""
        attrs = get_attributes_from_string(str_input)

        solution_entry = None
        if "doi" in attrs:  # case doi
            solution_entry = self._search_doi(attrs["doi"])
        else:
            coordinates = dict_to_coordinates(attrs)
            if "catalog" not in attrs:
                solution_entry = self._search_in_local_catalog(coordinates)  # resolve in local catalog first!

            if not solution_entry:
                if "catalog" in attrs:  # resolve in specific catalog
                    catalog_id = self.album.catalogs().get_by_name(attrs["catalog"]).catalog_id()
                    solution_entry = self._search_in_specific_catalog(catalog_id, coordinates)
                else:
                    solution_entries = self._search_in_catalogs(coordinates)  # resolve anywhere

                    if solution_entries and len(solution_entries) > 1:
                        module_logger().warning("Found several solutions... taking the first one! ")

                    if solution_entries:
                        solution_entry = solution_entries[0]

        return solution_entry

    def _search_by_coordinates(self, coordinates: ICoordinates) -> Optional[ICollectionIndex.ICollectionSolution]:
        solution_entry = self._search_in_local_catalog(coordinates)  # resolve in local catalog first!

        if not solution_entry:
            solution_entries = self._search_in_catalogs(coordinates)  # resolve anywhere

            if solution_entries and len(solution_entries) > 1:
                module_logger().warning("Found several solutions... taking the first one! ")

            if solution_entries:
                solution_entry = solution_entries[0]

        return solution_entry

    def _search_in_local_catalog(self, coordinates: ICoordinates) -> Optional[ICollectionIndex.ICollectionSolution]:
        """Searches in the local catalog only"""
        return self._search_in_specific_catalog(self.album.catalogs().get_local_catalog().catalog_id(), coordinates)

    def _search_in_specific_catalog(self, catalog_id, coordinates: ICoordinates) \
            -> Optional[ICollectionIndex.ICollectionSolution]:
        """Searches in a given catalog only"""
        return self.catalog_collection.get_solution_by_catalog_grp_name_version(catalog_id, coordinates)

    def _search_in_catalogs(self, coordinates: ICoordinates) -> List[ICollectionIndex.ICollectionSolution]:
        """Searches the whole collection giving coordinates"""
        solution_entries = self.catalog_collection.get_solutions_by_grp_name_version(coordinates)

        return solution_entries if solution_entries else None

    def retrieve_and_load_resolve_result(self, resolve_result: ICollectionSolution):
        if not Path(resolve_result.path()).exists():
            self.solution_handler.retrieve_solution(resolve_result.catalog(),
                                               dict_to_coordinates(resolve_result.database_entry().setup()))
        resolve_result.set_loaded_solution(self.album.state_manager().load(resolve_result.path()))
        resolve_result.set_coordinates(resolve_result.loaded_solution().coordinates())

    @staticmethod
    def write_version_to_yml(path, name, version) -> None:
        write_dict_to_json(path, {
            "catalog_collection_name": name,
            "catalog_collection_version": version
        })



import pkgutil
from datetime import datetime
from typing import Optional, List

from album.core.api.model.catalog_index import ICatalogIndex
from album.core.model.database import Database
from album.core.utils.operations.file_operations import get_dict_entry, write_dict_to_json
from album.core.utils.operations.solution_operations import get_solution_hash, serialize_json
from album.runner import album_logging
from album.runner.core.api.model.coordinates import ICoordinates

module_logger = album_logging.get_active_logger


class CatalogIndex(ICatalogIndex, Database):
    version = "0.1.0"

    def __init__(self, name, path):
        """Init routine.

        Args:
            name:
                The name of the catalog.
            path:
                The path to the index file.

        """
        self.name = name
        super().__init__(path)

    def create(self):
        data = pkgutil.get_data('album.core.schema', 'catalog_index_schema.sql')

        cursor = self.get_cursor()
        cursor.executescript(data.decode("utf-8"))

        self.update_name_version(self.name, self.version, close=False)

        # explicitly commit and close
        self.close_current_connection(commit=True)

    def is_empty(self, close=True):
        cursor = self.get_cursor()
        r = cursor.execute("SELECT * FROM solution").fetchone()

        if close:
            self.close_current_connection()

        return False if r else True

    # ### catalog_index ###

    def update_name_version(self, name, version, close=True):
        module_logger().debug("Update index name: \"%s\" and version: \"%s\"" % (name, version))

        curr_name = self.get_name()

        cursor = self.get_cursor()
        if curr_name:
            cursor.execute(
                "UPDATE catalog_index SET name=:name, version=:version WHERE name_id=:name_id",
                {"name_id": 1, "name": name, "version": version}
            )
        else:
            cursor.execute(
                "INSERT INTO catalog_index values (?, ?, ?)",
                (1, name, version)
            )

        if close:
            self.close_current_connection()

    def get_name(self, close=True):
        module_logger().debug("Get catalog name...")

        cursor = self.get_cursor()
        r = cursor.execute(
            "SELECT * FROM catalog_index"
        ).fetchone()

        cur_name = r["name"] if r else None

        if close:
            self.close_current_connection()

        return cur_name

    def get_version(self, close=True):
        module_logger().debug("Get index version...")

        cursor = self.get_cursor()
        r = cursor.execute(
            "SELECT * FROM catalog_index"
        ).fetchone()

        cur_version = r["version"] if r else None

        if close:
            self.close_current_connection()

        return cur_version

    def get_all_solutions(self, close=True):
        module_logger().debug("Retrieve all solutions...")

        cursor = self.get_cursor()
        r = cursor.execute(
            "SELECT * FROM solution",
            {}).fetchall()

        solutions = []
        if r:
            for s in r:
                solution = dict(s)
                self._append_metadata_to_solution_dict(solution)
                solutions.append(solution)

        if close:
            self.close_current_connection()

        return solutions

    @staticmethod
    def get_solution_column_keys():
        return ['group', 'name', 'description', 'version', 'album_api_version', 'album_version', 'license',
                'acknowledgement', 'title', 'timestamp', 'doi', 'changelog']

    def _insert_solution(self, solution_attrs, close=True) -> int:
        hash_val = get_solution_hash(solution_attrs, self.get_solution_column_keys())
        solution_id = self.next_id("solution")

        cursor = self.get_cursor()
        cursor.execute(
            'INSERT INTO solution values (?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?)',
            (
                solution_id,
                solution_attrs['group'],
                solution_attrs['name'],
                get_dict_entry(solution_attrs, 'title'),
                solution_attrs['version'],
                datetime.now().isoformat(),
                get_dict_entry(solution_attrs, 'description'),
                get_dict_entry(solution_attrs, 'doi'),  # allow to be none
                get_dict_entry(solution_attrs, 'license'),
                get_dict_entry(solution_attrs, 'album_version'),
                get_dict_entry(solution_attrs, 'album_api_version'),
                get_dict_entry(solution_attrs, 'changelog'),  # allow to be none
                get_dict_entry(solution_attrs, 'acknowledgement'),
                hash_val
            )
        )
        if 'authors' in solution_attrs:
            for author in solution_attrs['authors']:
                author_id = self._exists_author(author, close=False)
                if not author_id:
                    author_id = self._insert_author(author, close=False)
                solution_author_id = self.next_id('solution_author')
                cursor.execute(
                    'INSERT INTO solution_author values (?, ?, ?)',
                    (
                        solution_author_id,
                        solution_id,
                        author_id
                    )
                )

        if 'tags' in solution_attrs:
            for tag in solution_attrs['tags']:
                tag_id = self._exists_tag(tag, close=False)
                if not tag_id:
                    tag_id = self._insert_tag(tag, close=False)
                solution_tag_id = self.next_id('solution_tag')
                cursor.execute(
                    'INSERT INTO solution_tag values (?, ?, ?)',
                    (
                        solution_tag_id,
                        solution_id,
                        tag_id
                    )
                )

        if 'args' in solution_attrs:
            for argument in solution_attrs['args']:
                argument_id = self._exists_argument(argument, close=False)
                if not argument_id:
                    argument_id = self._insert_argument(argument, close=False)
                solution_argument_id = self.next_id('solution_argument')
                cursor.execute(
                    'INSERT INTO solution_argument values (?, ?, ?)',
                    (
                        solution_argument_id,
                        solution_id,
                        argument_id
                    )
                )
        if 'cite' in solution_attrs:
            for citation in solution_attrs['cite']:
                citation_id = self._exists_citation(citation, close=False)
                if not citation_id:
                    citation_id = self._insert_citation(citation, close=False)
                solution_citation_id = self.next_id('solution_citation')
                cursor.execute(
                    'INSERT INTO solution_citation values (?, ?, ?)',
                    (
                        solution_citation_id,
                        solution_id,
                        citation_id
                    )
                )

        if 'covers' in solution_attrs:
            for cover in solution_attrs['covers']:
                if not self._exists_cover(cover, solution_id):
                    self._insert_cover(cover, solution_id, close=False)

        if 'documentation' in solution_attrs:
            for documentation in solution_attrs['documentation']:
                if not self._exists_documentation(documentation, solution_id):
                    self._insert_documentation(documentation, solution_id, close=False)

        self.save()

        if close:
            self.close_current_connection()

        return solution_id

    def _exists_author(self, author_name, close=True):
        cursor = self.get_cursor()
        r = cursor.execute(
            "SELECT * FROM author WHERE name=:author_name",
            {
                "author_name": author_name
            }
        ).fetchone()

        if close:
            self.close_current_connection()

        return r["author_id"] if r else None

    def _insert_author(self, author, close=True):
        author_id = self.next_id("author")

        cursor = self.get_cursor()
        cursor.execute(
            "INSERT INTO author values (?, ?)",
            (
                author_id,
                author
            )
        )

        if close:
            self.close_current_connection()

        return author_id

    def _exists_tag(self, tag_name, close=True):
        cursor = self.get_cursor()
        r = cursor.execute(
            "SELECT * FROM tag WHERE name=:tag_name ",
            {
                "tag_name": tag_name
            }
        ).fetchone()

        if close:
            self.close_current_connection()

        return r["tag_id"] if r else None

    def _insert_tag(self, tag, close=True):
        tag_id = self.next_id("tag")

        cursor = self.get_cursor()
        cursor.execute(
            "INSERT INTO tag values (?, ?, ?)",
            (
                tag_id,
                tag,
                "manual"
            )
        )

        if close:
            self.close_current_connection()

        return tag_id

    def _exists_argument(self, argument, close=True):
        cursor = self.get_cursor()

        exc_str = "SELECT * FROM argument WHERE name=:argument_name AND description=:argument_description "
        exc_val = {
            "argument_name": argument["name"],
            "argument_description": argument["description"],
        }
        argument_type = get_dict_entry(argument, "type")
        argument_default_value = get_dict_entry(argument, "default_value")

        if argument_type:
            exc_str += "AND type=:argument_type "
            exc_val["argument_type"] = argument_type
        else:
            exc_str += "AND type IS NULL "

        if argument_default_value:
            exc_str += "AND default_value=:argument_default_value "
            exc_val["argument_default_value"] = argument_default_value
        else:
            exc_str += "AND default_value IS NULL "

        r = cursor.execute(exc_str, exc_val).fetchone()

        if close:
            self.close_current_connection()

        return r["argument_id"] if r else None

    def _insert_argument(self, argument, close=True):
        argument_id = self.next_id("argument")

        cursor = self.get_cursor()
        cursor.execute(
            "INSERT INTO argument values (?, ?, ?, ?, ?, ?)",
            (
                argument_id,
                argument["name"],
                get_dict_entry(argument, "type"),
                argument["description"],
                get_dict_entry(argument, "default_value"),
                get_dict_entry(argument, "required")
            )
        )

        if close:
            self.close_current_connection()

        return argument_id

    def _exists_citation(self, citation, close=True):
        cursor = self.get_cursor()

        exc_str = "SELECT * FROM citation WHERE text=:citation_text "
        exc_val = {
            "citation_text": citation["text"],
        }
        citation_doi = get_dict_entry(citation, "doi")

        if citation_doi:
            exc_str += "AND doi=:citation_doi"
            exc_val["citation_doi"] = citation_doi
        else:
            exc_str += "AND doi IS NULL"

        r = cursor.execute(exc_str, exc_val).fetchone()

        if close:
            self.close_current_connection()

        return r["citation_id"] if r else None

    def _insert_citation(self, citation, close=True):
        citation_id = self.next_id("citation")
        cursor = self.get_cursor()
        cursor.execute(
            "INSERT INTO citation values (?, ?, ?, ?)",
            (
                citation_id,
                citation["text"],
                get_dict_entry(citation, "doi"),
                get_dict_entry(citation, "url")
            )
        )

        if close:
            self.close_current_connection()

        return citation_id

    def _exists_cover(self, cover, solution_id, close=True):
        cursor = self.get_cursor()
        r = cursor.execute(
            "SELECT * FROM cover WHERE source=:cover_source AND description=:cover_description "
            "AND solution_id=:solution_id",
            {
                "cover_source": cover["source"],
                "cover_description": cover["description"],
                "solution_id": solution_id
            }
        ).fetchone()

        if close:
            self.close_current_connection()

        return r["cover_id"] if r else None

    def _insert_cover(self, cover, solution_id, close=True):
        cover_id = self.next_id("cover")

        cursor = self.get_cursor()
        cursor.execute(
            "INSERT INTO cover values (?, ?, ?, ?)",
            (
                cover_id,
                solution_id,
                cover["source"],
                cover["description"]
            )
        )

        if close:
            self.close_current_connection()

        return cover_id

    def _exists_documentation(self, documentation, solution_id, close=True):
        cursor = self.get_cursor()
        r = cursor.execute(
            "SELECT * FROM main.documentation WHERE documentation=:documentation "
            "AND solution_id=:solution_id",
            {
                "documentation": documentation,
                "solution_id": solution_id,
            }
        ).fetchone()

        if close:
            self.close_current_connection()

        return r["documentation_id"] if r else None

    def _insert_documentation(self, documentation, solution_id, close=True):
        documentation_id = self.next_id("documentation")
        cursor = self.get_cursor()
        cursor.execute(
            "INSERT INTO documentation values (?, ?, ?)",
            (
                documentation_id,
                solution_id,
                documentation
            )
        )

        if close:
            self.close_current_connection()

        return documentation_id

    def get_solution(self, solution_id, close=True) -> Optional[dict]:
        module_logger().debug("Get solution by id: \"%s\"..." % solution_id)

        cursor = self.get_cursor()
        r = cursor.execute(
            "SELECT * FROM solution WHERE solution_id=:solution_id",
            {
                "solution_id": solution_id,
            }).fetchone()
        solution = None
        if r:
            solution = dict(r)
            self._append_metadata_to_solution_dict(solution)

        if close:
            self.close_current_connection()

        return solution

    def get_solution_by_coordinates(self, coordinates: ICoordinates, close=True) -> Optional[dict]:
        module_logger().debug("Get solution by coordinates: \"%s\"..." % str(coordinates))

        cursor = self.get_cursor()
        r = cursor.execute(
            "SELECT s.* FROM solution s "
            "WHERE s.\"group\"=:group AND s.name=:name AND s.version=:version",
            {
                "group": coordinates.group(),
                "name": coordinates.name(),
                "version": coordinates.version(),
            }
        ).fetchone()

        solution = None
        if r:
            solution = dict(r)
            self._append_metadata_to_solution_dict(solution)

        if close:
            self.close_current_connection()

        return solution

    def _append_metadata_to_solution_dict(self, solution):
        solution_id = solution["solution_id"]
        solution["authors"] = self._get_authors_by_solution(solution_id)
        solution["covers"] = self._get_covers_by_solution(solution_id)
        solution["documentation"] = self._get_documentation_by_solution(solution_id)
        solution["args"] = self._get_arguments_by_solution(solution_id)
        solution["cite"] = self._get_citations_by_solution(solution_id)
        solution["tags"] = self._get_tags_by_solution(solution_id)

    def get_solution_by_doi(self, doi, close=True) -> Optional[dict]:
        module_logger().debug("Get solution by doi: \"%s\"..." % doi)

        cursor = self.get_cursor()
        r = cursor.execute(
            "SELECT * FROM solution WHERE doi=:doi_value",
            {
                "doi_value": doi,
            }
        ).fetchone()

        solution = None
        if r:
            solution = dict(r)

        if close:
            self.close_current_connection()

        return solution

    def get_all_solution_versions(self, group, name, close=True) -> Optional[List[dict]]:
        module_logger().debug("Get solution by group and name: \"%s:%s\"..." % (group, name))

        cursor = self.get_cursor()
        r = cursor.execute(
            "SELECT * FROM solution WHERE \"group\"=:group_value AND name=:name_value ORDER BY version DESC",
            {
                "group_value": group,
                "name_value": name,
            }
        ).fetchall()

        solutions = []
        if r:
            for solution in r:
                solutions.append(dict(solution))

        if close:
            self.close_current_connection()

        return solutions

    def _update_solution(self, coordinates: ICoordinates, solution_attrs: dict, close=True) -> None:
        # it is easier to delete and insert again instead of updating all connection-tables
        solution_dict = self.get_solution_by_coordinates(coordinates, close=False)
        if solution_dict:
            self.remove_solution(solution_dict["solution_id"])
        else:
            raise RuntimeError("Cannot update solution \"%s\"! Does not exist!" % str(coordinates))

        self._insert_solution(solution_attrs, close=False)

        self.save()

        if close:
            self.close_current_connection()

    def remove_solution(self, solution_id, close=True):
        cursor = self.get_cursor()

        cursor.execute(
            "DELETE FROM cover WHERE solution_id=:solution_id",
            {
                "solution_id": solution_id
            }
        )
        cursor.execute(
            "DELETE FROM documentation WHERE solution_id=:solution_id",
            {
                "solution_id": solution_id
            }
        )

        cursor.execute(
            "DELETE FROM solution_tag WHERE solution_id=:solution_id",
            {
                "solution_id": solution_id
            }
        )

        cursor.execute(
            "DELETE FROM solution_author WHERE solution_id=:solution_id",
            {
                "solution_id": solution_id
            }
        )

        cursor.execute(
            "DELETE FROM solution_citation WHERE solution_id=:solution_id",
            {
                "solution_id": solution_id
            }
        )

        cursor.execute(
            "DELETE FROM solution_argument WHERE solution_id=:solution_id",
            {
                "solution_id": solution_id
            }
        )

        cursor.execute(
            "DELETE FROM tag "
            "WHERE NOT EXISTS (SELECT st.tag_id FROM solution_tag st "
            "WHERE tag.tag_id = st.tag_id)")

        cursor.execute(
            "DELETE FROM argument "
            "WHERE NOT EXISTS (SELECT sa.argument_id FROM solution_argument sa "
            "WHERE argument.argument_id = sa.argument_id)")

        cursor.execute(
            "DELETE FROM citation "
            "WHERE NOT EXISTS (SELECT sc.citation_id FROM solution_citation sc "
            "WHERE citation.citation_id = sc.citation_id)")

        cursor.execute(
            "DELETE FROM author "
            "WHERE NOT EXISTS (SELECT sa.author_id FROM solution_author sa "
            "WHERE author.author_id = sa.author_id)")

        cursor.execute(
            "DELETE FROM solution WHERE solution_id=:solution_id",
            {
                "solution_id": solution_id
            }
        )

        self.save()

        if close:
            self.close_current_connection()

    def remove_solution_by_group_name_version(self, coordinates: ICoordinates, close=True):
        solution_dict = self.get_solution_by_coordinates(coordinates, close=close)
        if solution_dict:
            self.remove_solution(solution_dict["solution_id"])

    # ### catalog_features ###

    def update(self, coordinates: ICoordinates, solution_attrs: dict, close=True):
        if self.get_solution_by_coordinates(coordinates, close=False):
            module_logger().debug("Update solution...")
            self._update_solution(coordinates, solution_attrs, close=close)
        else:
            module_logger().debug("Insert solution...")
            self._insert_solution(solution_attrs, close=close)

    def save(self):
        module_logger().debug("Saving index...")
        self.get_connection().commit()

    def export(self, path, export_format="JSON", close=True):
        module_logger().debug("Export index...")

        cursor = self.get_cursor()
        r = cursor.execute("SELECT * FROM solution").fetchall()

        if export_format == "JSON":
            json_dumps_list = []
            for solution in r:
                solution = dict(solution)
                self._append_metadata_to_solution_dict(solution)
                json_dumps_list.append(serialize_json(solution))

            write_dict_to_json(path, json_dumps_list)
        else:
            raise NotImplementedError("Unsupported format \"%s\"" % export_format)

        if close:
            self.close_current_connection()

    def __len__(self, close=True):
        cursor = self.get_cursor()
        r = cursor.execute("SELECT COUNT(*) FROM solution").fetchone()
        r = r[0]

        if close:
            self.close_current_connection()

        return r

    def _get_authors_by_solution(self, solution_id, close=True):
        cursor = self.get_cursor()
        r = cursor.execute(
            "SELECT a.* FROM author a "
            "JOIN solution_author sa ON sa.author_id = a.author_id "
            "WHERE sa.solution_id=:solution_id",
            {
                "solution_id": solution_id
            }
        ).fetchall()

        res = []
        for row in r:
            res.append(row["name"])

        if close:
            self.close_current_connection()

        return res

    def _get_tags_by_solution(self, solution_id, close=True):
        cursor = self.get_cursor()
        r = cursor.execute(
            "SELECT t.* FROM tag t "
            "JOIN solution_tag st ON st.tag_id = t.tag_id "
            "WHERE st.solution_id=:solution_id",
            {
                "solution_id": solution_id
            }
        ).fetchall()

        res = []
        for row in r:
            res.append(row["name"])

        if close:
            self.close_current_connection()

        return res

    def _get_citations_by_solution(self, solution_id, close=True):
        cursor = self.get_cursor()
        r = cursor.execute(
            "SELECT c.* FROM citation c "
            "JOIN solution_citation sc ON sc.citation_id = c.citation_id "
            "WHERE sc.solution_id=:solution_id",
            {
                "solution_id": solution_id
            }
        ).fetchall()

        res = []
        for row in r:
            res.append(dict(row))

        if close:
            self.close_current_connection()

        return res

    def _get_arguments_by_solution(self, solution_id, close=True):
        cursor = self.get_cursor()
        r = cursor.execute(
            "SELECT a.* FROM argument a "
            "JOIN solution_argument sa ON sa.argument_id = a.argument_id "
            "WHERE sa.solution_id=:solution_id",
            {
                "solution_id": solution_id
            }
        ).fetchall()

        res = []
        for row in r:
            res.append(dict(row))

        if close:
            self.close_current_connection()

        return res

    def _get_covers_by_solution(self, solution_id, close=True):
        cursor = self.get_cursor()
        r = cursor.execute(
            "SELECT c.* FROM cover c "
            "WHERE c.solution_id=:solution_id",
            {
                "solution_id": solution_id
            }
        ).fetchall()

        res = []
        for row in r:
            res.append(dict(row))

        if close:
            self.close_current_connection()

        return res

    def _get_documentation_by_solution(self, solution_id, close=True):
        cursor = self.get_cursor()
        r = cursor.execute(
            "SELECT d.* FROM documentation d "
            "WHERE d.solution_id=:solution_id",
            {
                "solution_id": solution_id
            }
        ).fetchall()

        res = []
        for row in r:
            res.append(row["documentation"])

        if close:
            self.close_current_connection()

        return res



import re
import tempfile
from pathlib import Path

import requests
from requests.adapters import HTTPAdapter
from urllib3.util import Retry

from album.ci.utils.zenodo_api import ResponseStatus
from album.core.utils.operations.file_operations import create_path_recursively
from album.runner import album_logging

module_logger = album_logging.get_active_logger


def is_downloadable(url):
    """Shows if url is a downloadable resource."""
    with _get_session() as s:
        h = s.head(url, allow_redirects=True)
        header = h.headers
        content_type = header.get('content-type')
        if 'html' in content_type.lower():
            return False
        return True


def download_resource(url, path):
    """Downloads a resource given its url."""
    module_logger().debug("Download url %s to %s..." % (url, path))

    path = Path(path)

    if not is_downloadable(url):
        raise AssertionError("Resource \"%s\" not downloadable!" % url)

    r = _request_get(url)

    create_path_recursively(path.parent)
    with open(path, "wb") as f:
        for chunk in r:
            f.write(chunk)

    return path


def _get_session():
    s = requests.Session()
    retry = Retry(connect=3, backoff_factor=0.5)

    adapter = HTTPAdapter(max_retries=retry)

    s.mount('http://', adapter)
    s.mount('https://', adapter)

    return s


def _request_get(url):
    """Get a response from a request to a resource url."""
    with _get_session() as s:
        r = s.get(url, allow_redirects=True, stream=True)

        if r.status_code != ResponseStatus.OK.value:
            raise ConnectionError("Could not connect to resource %s!" % url)

        return r


def retrieve_redirect_url(url):
    with _get_session() as s:
        r = s.get(url, allow_redirects=True, stream=False)

        if r.status_code != ResponseStatus.OK.value:
            raise ConnectionError("Could not connect to resource %s!" % url)

        return r.url


def is_url(str_input: str):
    """Parses a url."""
    url_regex = re.compile(
        r'^(?:http|ftp)s?://'  # http:// or https://
        r'(?:(?:[A-Z0-9](?:[A-Z0-9-]{0,61}[A-Z0-9])?\.)+(?:[A-Z]{2,6}\.?|[A-Z0-9-]{2,}\.?)|'  # domain...
        r'localhost|'  # localhost...
        r'\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3})'  # ...or ip
        r'(?::\d+)?'  # optional port
        r'(?:/?|[/?]\S+)$', re.IGNORECASE)
    return re.match(url_regex, str_input) is not None


def download(str_input, base):
    """Downloads a solution file into a temporary file."""
    Path(base).mkdir(exist_ok=True, parents=True)

    with _get_session() as s:
        r = s.get(str_input, allow_redirects=True)

        content_type = r.headers.get('content-type').lower()
        if content_type == 'application/zip':
            suffix = '.zip'
            new_file, tmp_file_name = tempfile.mkstemp(dir=base, suffix=suffix)
        else:
            new_file, tmp_file_name = tempfile.mkstemp(dir=base)
        with open(tmp_file_name, 'wb') as out:
            out.write(r.content)
        return Path(tmp_file_name)



import errno
import json
import os
import platform
import random
import re
import shutil
import stat
import sys
import tempfile
from pathlib import Path
from typing import Optional
from zipfile import ZipFile

import yaml

from album.runner import album_logging

module_logger = album_logging.get_active_logger
enc = sys.getfilesystemencoding()


def get_line_indent(line):
    """Returns the line indent of a line."""
    return int(len(re.findall('^[ ]*', line)[0]) / 4)


def indent_to_space(indent: int):
    """Returns the right number of spaces belonging to an indent"""
    return ''.join([' '] * 4 * indent)


def is_comment(line):
    """Finds out if line is a comment line."""
    if re.match('^[ ]*[#]', line):
        return True
    return False


def is_blank_line(line):
    """Finds out if line is blank."""
    if re.match('^\n', line):
        return True
    return False


def to_executable(argument_parsing):
    """Connects all arguments in a argument list"""
    executable = ""
    for arg in argument_parsing:
        executable += arg
    return executable


def get_dict_from_yml(yml_file):
    """Reads a dictionary from a file in yml format."""
    with open(yml_file, 'r') as yml_f:
        d = yaml.safe_load(yml_f)

    if not isinstance(d, dict):
        raise TypeError("Yaml file %s invalid!" % str(yml_file))

    return d


def write_dict_to_yml(yml_file, d):
    """Writes a dictionary to a file in yml format."""
    yml_file = Path(yml_file)
    create_path_recursively(yml_file.parent)

    with open(yml_file, 'w+') as yml_f:
        yml_f.write(yaml.dump(d, Dumper=yaml.Dumper))

    return True


def get_dict_entry(d, key, allow_none=True, message=None):
    """Receive an entry from a dictionary.

    Args:
        d:
            The yml dictionary.
        key:
            The key.
        allow_none:
            boolean flag indicating whether to allow key to exist or not.
        message:
            When allow_none false this is the message printed in the exception.


    Returns:
        The value or None

    Raises:
        KeyError if allow_none is False and key not found.

    """
    try:
        val = d[key]
    except KeyError:
        val = None
        if not allow_none:
            raise KeyError(message)

    return val


def write_dict_to_json(json_file, d):
    """Writes dictionary to JSON file."""
    json_file = Path(json_file)
    create_path_recursively(json_file.parent)

    with open(json_file, 'w+') as json_f:
        json_f.write(json.dumps(d))

    return True


def get_dict_from_json(json_file):
    """Reads dictionary from JSON file."""
    json_file = Path(json_file)

    with open(json_file, 'r') as json_f:
        d = json.load(json_f)

    return d


def folder_empty(path) -> bool:
    path = Path(path)

    if path.exists():
        if path.is_dir() and os.listdir(path) == []:
            return True
        return False
    return True


def create_empty_file_recursively(path_to_file):
    """Creates an empty file. Creates missing parent folders."""
    p = Path(path_to_file)
    p.parent.mkdir(parents=True, exist_ok=True)
    p.touch(exist_ok=True)

    return True


def create_path_recursively(path):
    """Creates a path. Creates missing parent folders."""
    p = Path(path)
    p.mkdir(parents=True, exist_ok=True)

    return True


def create_paths_recursively(paths):
    """Create paths. Creates missing parent folders."""
    for p in paths:
        create_path_recursively(p)


def copy_folder(folder_to_copy, destination, copy_root_folder=True, force_copy=False):
    folder_to_copy = Path(folder_to_copy)
    destination = Path(destination)

    if os.path.exists(destination) and os.path.samefile(folder_to_copy, destination):
        return destination

    if copy_root_folder:
        destination = destination.joinpath(folder_to_copy.name)

    if force_copy:
        force_remove(destination)

    create_path_recursively(destination)

    for root, dirs, files in os.walk(folder_to_copy):
        root = Path(root)

        for d in dirs:
            copy_folder(root.joinpath(d), destination.joinpath(d), copy_root_folder=False)
        for fi in files:
            copy(root.joinpath(fi), destination)
        break

    return destination


def copy(file, path_to) -> Path:
    """Copies a file A to either folder B or file B. Makes sure folder structure for target exists."""
    file = Path(file)
    path_to = Path(path_to)

    if os.path.exists(path_to) and os.path.samefile(file, path_to):
        return path_to

    create_path_recursively(path_to.parent)

    return Path(shutil.copy(file, path_to))


def copy_in_file(file_content, file_path) -> Path:
    """Copies a stream content to a file."""
    file_path = Path(file_path)
    create_path_recursively(file_path.parent)
    if file_path.is_file():
        file_path.unlink()
    with open(file_path, "w") as script_file:
        script_file.write(file_content)
    return file_path


def zip_folder(folder_path, zip_archive_file):
    """Creates a zip archive of a given folder."""
    zip_archive_file = Path(zip_archive_file)

    if zip_archive_file.suffix == ".zip":  # remove suffix as it appends automatically
        zip_archive_file = zip_archive_file.with_suffix("")

    create_path_recursively(zip_archive_file.parent)

    return shutil.make_archive(str(zip_archive_file), 'zip', folder_path)


def zip_paths(paths_to_include, zip_archive_file, tmp_dir=None):
    """Creates a zip archive including all given files. Copies the data into a tmp directory first."""

    with tempfile.TemporaryDirectory(dir=tmp_dir) as tmp_folder:
        for file in paths_to_include:
            file = Path(file)
            if file.is_file():
                copy(file, tmp_folder)
            elif file.is_dir():
                copy_folder(file, tmp_folder)
            else:
                raise FileNotFoundError("Could not find file %s" % str(file))

        return zip_folder(tmp_folder, zip_archive_file)


def unzip_archive(zip_archive_file, target_folder=None):
    """Unzips a archive file to a given folder or the folder where it is located (default)."""
    zip_archive_file = Path(zip_archive_file)
    target_folder = Path(target_folder) if target_folder else zip_archive_file.parent

    create_path_recursively(target_folder)

    shutil.unpack_archive(str(zip_archive_file), str(target_folder))

    return target_folder

def force_remove(path, warning=True):
    path = Path(path)
    if path.exists():
        try:
            if path.is_file():
                try:
                    path.unlink()
                except PermissionError:
                    handle_remove_readonly(os.unlink, path, sys.exc_info())
            else:
                shutil.rmtree(str(path), ignore_errors=False, onerror=handle_remove_readonly)
        except PermissionError as e:
            module_logger().warn("Cannot delete %s." % str(path))
            if not warning:
                raise e


def remove_link(link_target):
    if link_target:
        dispose_op = getattr(link_target, 'dispose', None)
        if callable(dispose_op):
            dispose_op()
        else:
            raise RuntimeError("Path doesn't seem to be a link")
        force_remove(link_target)


def handle_remove_readonly(func, path, exc):
    excvalue = exc[1]
    if func in (os.rmdir, os.remove, os.unlink) and excvalue.errno == errno.EACCES:
        os.chmod(path, stat.S_IRWXU | stat.S_IRWXG | stat.S_IRWXO)  # 0777
        func(path)
    else:
        raise


def rand_folder_name(f_len=8):
    characters = "abcdefghijklmnopqrstuvwxyz0123456789_"
    s = []
    for i in range(0, f_len):
        s.append(characters[random.randint(0, len(characters) - 1)])
    return "".join(s)


def check_zip(path):
    return not ZipFile(path).testzip()


def construct_cache_link_target(lnk_path: Path, link, target, create=True) -> Optional[Path]:
    operation_system = platform.system().lower()
    root = lnk_path.joinpath(target)
    if 'windows' in operation_system:
        from pylnk3 import Lnk, for_file
        link = str(link) + '.lnk'
        if Path(link).exists():
            resolve = Path(Lnk(link).path).absolute()
            return resolve
        if create:
            i = 0
            while root.joinpath("%s" % i).exists():
                i += 1
            target = root.joinpath('%s' % i)
            target.mkdir(exist_ok=True, parents=True)
            Path(link).parent.mkdir(parents=True, exist_ok=True)
            for_file(
                target_file=os.path.normpath(target),
                lnk_name=link
            )
            resolve = target.absolute()
            return resolve
    else:
        if os.path.islink(link):
            return Path(link).resolve()
        if create:
            i = 0
            while root.joinpath("%s" % i).exists():
                i += 1
            target = root.joinpath('%s' % i)
            target.mkdir(exist_ok=True, parents=True)
            Path(link).parent.mkdir(parents=True, exist_ok=True)

            os.symlink(target, link, target_is_directory=True)
            return target.resolve()
    return None

def get_link_target(link):
    operation_system = platform.system().lower()
    if 'windows' in operation_system:
        from pylnk3 import Lnk, for_file
        link = str(link) + '.lnk'
        if Path(link).exists():
            resolve = Path(Lnk(link).path).absolute()
            return resolve
    else:
        if Path(link).exists():
            return Path(link).resolve()
    return None



from album.ci.controller.release_manager import ReleaseManager


def configure_repo(release_manager: ReleaseManager, args):
    release_manager.configure_repo(args.ci_user_name, args.ci_user_email)


def configure_ssh(release_manager: ReleaseManager, args):
    release_manager.configure_ssh(args.ci_project_path)


def zenodo_publish(release_manager: ReleaseManager, args):
    release_manager.zenodo_publish(args.branch_name, args.zenodo_base_url, args.zenodo_access_token)


def zenodo_upload(release_manager: ReleaseManager, args):
    release_manager.zenodo_upload(args.branch_name, args.zenodo_base_url, args.zenodo_access_token, args.report_file)


def update_index(release_manager: ReleaseManager, args):
    release_manager.update_index(args.branch_name, args.doi, args.deposit_id)


def commit_changes(release_manager: ReleaseManager, args):
    release_manager.commit_changes(
        args.branch_name, args.ci_user_name, args.ci_user_email
    )


def merge(release_manager: ReleaseManager, args):
    release_manager.merge(args.branch_name, args.dry_run, args.push_option, args.ci_user_name, args.ci_user_email)



import argparse
import sys
import traceback

from album import core
from album.api import Album
from album.commandline import add_catalog, remove_catalog, deploy, \
    install, repl, run, search, start_server, test, update, clone, upgrade, index, uninstall, info
from album.core.utils.subcommand import SubProcessError
from album.runner.album_logging import debug_settings, get_active_logger, LogLevel, to_loglevel


def main():
    """Entry points of `album`."""

    parser = create_parser()

    get_active_logger().debug('Parsing base album call arguments...')
    args = parser.parse_known_args()
    __handle_args(args, parser)


def __handle_args(args, parser):
    """Handles all arguments provided after the album command."""
    level = args[0].log
    print_json = getattr(args[0], "json", False)
    __run_subcommand(args, parser, level, print_json)


def _capture_output():
    logger = get_active_logger()
    logger.handlers.clear()


def _handle_exception(e, silent: bool = False, exit_status=None):
    # if not silent:
    get_active_logger().error('album command failed: %s' % str(e))
    get_active_logger().debug(traceback.format_exc())
    sys.exit(e)
    # else:
    #     sys.exit(exit_status)


def __run_subcommand(args, parser, level: LogLevel, print_json):
    """Calls a specific album subcommand."""
    album_command = ""
    try:
        album_command = sys.argv[1]  # album command always expected at second position
    except IndexError:
        parser.error("Please provide a valid action!")
    get_active_logger().debug("Running %s subcommand..." % album_command)
    sys.argv = [sys.argv[0]] + args[1]

    # Makes sure album is initialized.
    album_instance = create_album_instance(level)
    if print_json:
        _capture_output()
    get_active_logger().info("album version %s | contact via %s " % (core.__version__, core.__email__))
    album_instance.load_or_create_collection()

    try:
        args[0].func(album_instance, args[0])  # execute entry point function
    except SubProcessError as e:
        _handle_exception(e, silent=True, exit_status=e.exit_status)
    except Exception as e:
        _handle_exception(e)


def create_album_instance(level: LogLevel) -> Album:
    return Album.Builder().log_level(level).build()


def create_parser():
    """Creates a parser for all known album arguments."""
    parser = AlbumParser()
    p = parser.create_command_parser('search', search, 'search for an album solution using keywords.')
    p.add_argument('keywords', type=str, nargs='+', help='Search keywords')
    p = parser.create_file_command_parser('run', run, 'run an album solution.')
    p.add_argument(
        '--run-immediately',
        required=False,
        help='When the solution to run consists of several steps, indicates whether to immediately run '
             'a step or to wait for all steps to be prepared to run.',
        default=False,
        action='store_true')
    parser.create_file_command_parser('repl', repl, 'get an interactive repl for an album solution.')
    p = parser.create_file_command_parser('deploy', deploy, 'deploy an album solution.')
    p.add_argument(
        '--dry-run',
        required=False,
        help='Parameter to indicate a dry run and only show what would happen.',
        action='store_true'
    )
    p.add_argument(
        '--catalog',
        required=False,
        help='Specify a catalog name to deploy to. Must be configured! '
             '\"catalog_local\" refers to the local catalog. Default is None.',
        default=None)
    p.add_argument(
        '--push-option',
        required=False,
        help='Push options for the catalog repository.',
        default=None,
        nargs="+"
    )
    p.add_argument(
        '--git-email',
        required=False,
        help='Email to use for all git operations. If none given, system is required to be proper configured!',
        default=None
    )
    p.add_argument(
        '--git-name',
        required=False,
        help='Name to use for all git operations. If none given, system is required to be proper configured!',
        default=None
    )
    p.add_argument(
        '--force-deploy',
        required=False,
        help='When specified,force deploys a solution to a catalog.'
             ' Useful if the solution has already been deployed once.',
        action='store_false'
    )
    p.add_argument(
        '--changelog',
        required=False,
        help='Description of changes from the previous version of the solution to this version.',
        default=None
    )

    parser.create_file_command_parser('install', install, 'install an album solution.')
    p = parser.create_file_command_parser('uninstall', uninstall, 'uninstall an album solution.')
    p.add_argument(
        '--uninstall-deps',
        required=False,
        help='Boolean to additionally remove all album dependencies. Choose between %s.'
             % ", ".join([str(True), str(False)]),
        default=False,
        action='store_true'
    )
    parser.create_file_command_parser('info', info, 'print information about an album solution.')
    parser.create_catalog_command_parser(
        'add-catalog', add_catalog,
        'add a catalog to your local album configuration file.'
    )
    parser.create_catalog_command_parser(
        'remove-catalog', remove_catalog,
        'remove a catalog from your local album configuration file.'
    )
    p = parser.create_command_parser(
        'update',
        update,
        'Update the catalog index files. Either all catalogs configured, or a specific one.'
    )
    p.add_argument('src', type=str, help='src of the catalog', nargs='?')
    p = parser.create_command_parser(
        'upgrade',
        upgrade,
        'upgrade the local collection from the catalog index files. Either all catalogs configured, or a specific one.'
    )
    p.add_argument('src', type=str, help='src of the catalog', nargs='?')
    p.add_argument(
        '--dry-run',
        required=False,
        help='Parameter to indicate a dry run and only show what would happen.',
        action='store_true'
    )
    p = parser.create_command_parser('clone', clone, 'clone an album solution or catalog template.')
    p.add_argument(
        'src',
        type=str,
        help='path for the solution file, group:name:version or name of the catalog template.'
    )
    p.add_argument(
        'target_dir',
        type=str,
        help='The target directory where the solution or catalog will be added to.'
    )
    p.add_argument(
        'name',
        type=str,
        help='The new name of the cloned solution or catalog.'
    )
    parser.create_command_parser('index', index, 'print the index of the local album collection.')
    parser.create_file_command_parser(
        'test', test, 'execute the test routine of a solution.')
    p = parser.create_command_parser('server', start_server,
                                     'start an album server.')
    p.add_argument('--port', type=int, required=False, default=8080, help='Port')
    p.add_argument('--host', type=str, required=False, default="127.0.0.1", help='Host')
    return parser.parser


class ArgumentParser(argparse.ArgumentParser):
    """Override default error method of all parsers to show help of subcommand"""

    def error(self, message):
        self.print_help()
        self.exit(2, '%s: error: %s\n' % (self.prog, message))


class AlbumParser(ArgumentParser):
    def __init__(self):
        super().__init__()
        self.parent_parser = self.create_parent_parser()
        self.parser = self.create_parser()
        self.subparsers = self.parser.add_subparsers(title='actions', help='sub-command help')

    @staticmethod
    def create_parent_parser():
        """Parent parser for all subparsers to have the same set of arguments."""
        parent_parser = ArgumentParser(add_help=False)
        # parse logging
        parent_parser.add_argument(
            '--log',
            required=False,
            help='Logging level for your album command. Choose between %s' %
                 ", ".join([loglevel.name for loglevel in LogLevel]),
            default=LogLevel(debug_settings()),
            type=(lambda choice: to_loglevel(choice)),
        )
        parent_parser.add_argument(
            '--json',
            required=False,
            help='Adding this parameter prevents the log from being printed to the console.'
                 ' Instead, the result of the command - if present - is printed as JSON.',
            action='store_true'
        )
        return parent_parser

    def create_parser(self):
        """Creates the main parser for the album framework."""
        parser = ArgumentParser(
            add_help=True,
            description='album for running, building, and deploying computational solutions',
            parents=[self.parent_parser])
        return parser

    def create_command_parser(self, command_name, command_function, command_help):
        """Creates a parser for a album command, specified by a name, a function and a help description."""
        parser = self.subparsers.add_parser(command_name, help=command_help, parents=[self.parent_parser])
        parser.set_defaults(func=command_function)
        return parser

    def create_file_command_parser(self, command_name, command_function, command_help):
        """Creates a parser for a album command dealing with an album file.

        Parser is specified by a name, a function and a help description.
        """
        parser = self.create_command_parser(command_name, command_function, command_help)
        parser.add_argument('path', type=str, help='path for the solution file')
        return parser

    def create_catalog_command_parser(self, command_name, command_function, command_help):
        """Creates a parser for a album command dealing with an album catalog.

        Parser is specified by a name, a function and a help description.
        """
        parser = self.create_command_parser(command_name, command_function, command_help)
        parser.add_argument('src', type=str, help='src of the catalog')
        return parser


if __name__ == "__main__":
    main()
